# This program is public domain.
"""
Optical matrix form of the reflectivity calculation.

O.S. Heavens, Optical Properties of Thin Solid Films

This is a pure python implementation of reflectometry provided for
convenience when a compiler is not available.  The refl1d
application uses reflmodule to compute reflectivity.
"""
from numpy import asarray, isscalar, empty, ones, ones_like
from numpy import sqrt, exp, pi

def refl(kz, depth, rho, irho=0, sigma=0, rho_index=None):
    """
    Reflectometry as a function of kz for a set of slabs.

    :Parameters:
        *kz* : float[n] | inv angstrom
            Scattering vector $2*\pi*\sin(\theta)/\lambda$. This is $Q_z/2$.
        *depth* :  float[m] | |Ang|
            thickness of each layer.  The thickness of the incident medium
            and substrate are ignored.
        *rho*, *irho* :  float[n,k] | |1e-6/Ang^2|
            real and imaginary scattering length density for each layer for each kz
            Note: absorption cross section mu = 2 irho/lambda
        *sigma* : float[m-1] | |Ang|
            interfacial roughness.  This is the roughness between a layer
            and the subsequent layer.  There is no interface associated
            with the substrate.  The sigma array should have at least m-1
            entries, though it may have m with the last entry ignored.
        *rho_index* : int[m]
            index into rho vector for each kz

    Slabs are ordered with the surface SLD at index 0 and substrate at
    index -1, or reversed if kz < 0.
    """
    if isscalar(kz): kz = asarray([kz], 'd')

    m = len(depth)

    # Make everything into arrays
    depth = asarray(depth,'d')
    rho = asarray(rho,'d')
    irho = irho*ones_like(rho) if isscalar(irho) else asarray(irho,'d')
    sigma = sigma*ones(m-1,'d') if isscalar(sigma) else asarray(sigma,'d')

    # Repeat rho,irho columns as needed
    if rho_index != None:
        rho = rho[rho_index, :]
        irho = irho[rho_index, :]
    elif len(rho.shape) == 1:
        rho = rho[None,:]
        irho = irho[None,:]

    ## For kz < 0 we need to reverse the order of the layers
    ## Note that the interface array sigma is conceptually one
    ## shorter than rho,mu so when reversing it, start at n-1.
    ## This allows the caller to provide an array of length n
    ## corresponding to rho,mu or of length n-1.
    r = empty(len(kz),'D')
    r[kz>=1e-10] = calc(kz[kz>=1e-10], depth, rho, irho, sigma)
    r[kz<=1e-10] = calc(-kz[kz<=1e-10], depth[::-1], rho[:,::-1],
                        irho[:,::-1], sigma[m-2::-1])
    r[abs(kz) < 1e-10] = -1
    return r


def calc(kz, depth, rho, irho, sigma):
    if len(kz) == 0: return kz

    # Complex index of refraction is relative to the incident medium.
    # We can get the same effect using kz_rel^2 = kz^2 + 4*pi*rho_o
    # in place of kz^2, and ignoring rho_o
    kz_sq = kz**2 + 4e-6*pi*rho[:,0]
    k = kz

    # According to Heavens, the initial matrix should be [ 1 F; F 1],
    # which we do by setting B=I and M0 to [1 F; F 1].  An extra matrix
    # multiply versus some coding convenience.
    B11 = 1
    B22 = 1
    B21 = 0
    B12 = 0
    for i in range(0, len(depth)-1):
        k_next = sqrt(kz_sq - 4e-6*pi*(rho[:,i+1] + 1j*irho[:,i+1]))
        F = (k - k_next) / (k + k_next)
        F *= exp(-2*k*k_next*sigma[i]**2)
        #print "==== layer",i
        #print "kz:", kz
        #print "k:", k
        #print "k_next:",k_next
        #print "F:",F
        #print "rho:",rho[:,i+1]
        #print "irho:",irho[:,i+1]
        #print "d:",depth[i],"sigma:",sigma[i]
        M11 = exp(1j*k*depth[i]) if i>0 else 1
        M22 = exp(-1j*k*depth[i]) if i>0 else 1
        M21 = F*M11
        M12 = F*M22
        C1 = B11*M11 + B21*M12
        C2 = B11*M21 + B21*M22
        B11 = C1
        B21 = C2
        C1 = B12*M11 + B22*M12
        C2 = B12*M21 + B22*M22
        B12 = C1
        B22 = C2
        k = k_next

    r = B12/B11
    return r



"""
Reflectivity plugin for fitting GUI.

Note that the fitting infrastructure is still heavily tied to the reflectivity
modeling program, and this represents only the first tiny steps to separating
the two.
"""

__all__ = [ "data_view", "model_view", "new_model",
            "calc_errors", "show_errors" ]

import numpy
from bumps.fitproblem import FitProblem

from . import names as refl
from .errors import calc_errors, show_errors

# These are names used by the driver
def data_view():
    from .view.data_view import DataView
    return DataView

def model_view():
    from .view.model_view import ModelView
    return ModelView

def load_model(filename):
    if (filename.endswith('.so') or filename.endswith('.dll')
        or filename.endswith('.dyld')):
        from . import garefl
        problem = garefl.load(filename)
        return problem
    elif filename.endswith('.staj') or filename.endswith('.sta'):
        from .stajconvert import load_mlayer
        return FitProblem(load_mlayer(filename))
        #fit_all(problem.fitness, pmp=20)
    else:
        return None

def new_model():
    stack = refl.silicon(0,10) | refl.air
    instrument = refl.NCNR.NG1()
    probe = instrument.probe(T=numpy.linspace(0,5,200),
                             Tlo=0.2, slits_at_Tlo=2)
    M = refl.Experiment(sample=stack, probe=probe)
    problem = refl.FitProblem(M)
    return problem



# This program is in the public domain
# Author: Paul Kienzle
r"""
Experimental probe.

The experimental probe describes the incoming beam for the experiment.
Scattering properties of the sample are dependent on the type and
energy of the radiation.

See :ref:`data-guide` for details.

"""

# TOF stitching introduces a lot of complexity.
# Theta offset:
#   Q1 = 4 pi sin(T1 + offset)/L1
#   Q2 = 4 pi sin(T2 + offset)/L2
#   at offset=0,
#      Q1=Q2
#   at offset!=0,
#      Q1' ~ Q1 + 4pi offset/L1
#      Q2' ~ Q2 + 4pi offset/L2
#      => Q1' != Q2'
# Thick layers:
#   Since a given Q,dQ has multiple T,dT,L,dL, oversampling is going
#   to be very complicated.
# Energy dependent SLD:
#   Just because two points are at the same Q does not mean they have
#   the same theory function when scattering length density is
#   energy dependent
#
# Not stitching has its own issues.
# Calculation speed:
#   overlapping points are recalculated
#   profiles are recalculated
#
# Unstitched seems like the better bet.

from __future__ import with_statement, division
import os

import numpy
from numpy import sqrt, pi, inf, sign, log

from periodictable import nsf, xsf
from bumps.parameter import Parameter, Constant
from bumps.plotutil import coordinated_colors, auto_shift

from . import fresnel
from .material import Vacuum
from .resolution import QL2T, TL2Q, dTdL2dQ
from .stitch import stitch
from .reflectivity import convolve

PROBE_KW = ('T', 'dT', 'L', 'dL', 'data', 'name', 'filename',
            'intensity', 'background', 'back_absorption',
            'theta_offset', 'back_reflectivity', 'data')

def make_probe(**kw):
    """
    Return a reflectometry measurement object of the given resolution.
    """
    radiation = kw.pop('radiation')
    kw = dict((k,v) for k,v in kw.items() if k in PROBE_KW)
    if radiation == 'neutron':
        return NeutronProbe(**kw)
    else:
        return XrayProbe(**kw)


class Probe(object):
    r"""
    Defines the incident beam used to study the material.

    For calculation purposes, probe needs to return the values $Q_\text{calc}$
    at which the model is evaluated.  This is normally going to be the measured
    points only, but for some systems, such as those with very thick layers,
    oversampling is needed to avoid aliasing effects.

    A measurement point consists of incident angle, angular resolution, 
    incident wavelength, FWHM wavelength resolution, reflectivity and 
    uncertainty in reflectivity.

    A probe is a set of points, defined by vectors for point attribute.  For
    convenience, the attribute can be initialized with a scalar if it is
    constant throughout the measurement, but will be set to a vector in
    the probe.  The attributes are initialized as follows:

        *T* : float or [float] | degrees
            Incident angle
        *dT* : float or [float] | degrees
            FWHM angular resolution
        *L* : float or [float] | |Ang|
            Incident wavelength
        *dL* : float or [float] | |Ang|
            FWHM wavelength dispersion
        *data* : ([float],[float])
            R,dR reflectivity measurement and uncertainty

    Measurement properties:

        *intensity* : float or Parameter
           Beam intensity
        *background* : float or Parameter
           Constant background
        *back_absorption* : float or Parameter
           Absorption through the substrate relative to beam intensity
        *theta_offset* : float or Parameter
           Offset of the sample from perfect alignment
        *back_reflectivity* : True or False
           True if the beam enters through the substrate

    Measurement properties are fittable parameters.  *theta_offset* in
    particular should be set using *probe.theta_offset.dev(dT)*, with *dT*
    equal to the FWHM uncertainty in the peak position for the rocking curve,
    as measured in radians.  Changes to *theta_offset* will then be penalized
    in the cost function for the fit as if it were another measurement.  Use
    :meth:`alignment_uncertainty` to compute dT from the shape of the
    rocking curve.

    *intensity* and *back_absorption* are generally not needed --- scaling
    the reflected signal by an appropriate intensity measurement will correct
    for both of these during reduction.  *background* may be needed,
    particularly for samples with significant hydrogen content due to its
    large isotropic incoherent scattering cross section.

    View properties:

        *view* : string
            One of 'fresnel', 'logfresnel', 'log', 'linear', 'q4', 'residuals'
        *plot_shift* : float
            The number of pixels to shift each new dataset so
            datasets can be seen separately
        *residuals_shift* : 
            The number of pixels to shift each new set of residuals
            so the residuals plots can be seen separately.

    Normally *view* is set directly in the class rather than the
    instance since it is not specific to the view.  Fresnel and Q4 
    views are corrected for background and intensity; log and 
    linear views show the uncorrected data.
    """
    polarized = False
    Aguide = 270  # default guide field for unpolarized measurements
    view = "fresnel"
    plot_shift = 0
    residuals_shift = 0
    def __init__(self, T=None, dT=0, L=None, dL=0, data=None, 
                 intensity=1, background=0, back_absorption=1, theta_offset=0,
                 back_reflectivity=False, name=None, filename=None):
        if T is None or L is None:
            raise TypeError("T and L required")
        if not name and filename: 
            name = os.path.splitext(os.path.basename(filename))[0]
        qualifier = " "+name if name is not None else ""
        self.intensity = Parameter.default(intensity,name="intensity"+qualifier)
        self.background = Parameter.default(background,name="background"+qualifier,
                                            limits=[0,inf])
        self.back_absorption = Parameter.default(back_absorption,
                                                 name="back_absorption"+qualifier,
                                                 limits=[0,1])
        self.theta_offset = Parameter.default(theta_offset,name="theta_offset"+qualifier)

        self.back_reflectivity = back_reflectivity
        if data is not None:
            R,dR = data
        else:
            R,dR = None,None

        self._set_TLR(T,dT,L,dL,R,dR)
        self.name = name
        self.filename = filename

    def _set_TLR(self, T,dT,L,dL,R,dR):
        #if L is None:
        #    L = xsf.xray_wavelength(E)
        #    dL = L * dE/E
        #else:
        #    E = xsf.xray_energy(L)
        #    dE = E * dL/L

        Q = TL2Q(T=T, L=L)
        dQ = dTdL2dQ(T=T,dT=dT,L=L,dL=dL)

        # Make sure that we are dealing with vectors
        T,dT,L,dL = [numpy.ones_like(Q)*v for v in (T,dT,L,dL)]

        # Probe stores sorted values for convenience of resolution calculator
        idx = numpy.argsort(Q)
        self.T, self.dT = T[idx],dT[idx]
        self.L, self.dL = L[idx],dL[idx]
        self.Qo, self.dQ = Q[idx],dQ[idx]
        if R is not None: R = R[idx]
        if dR is not None: dR = dR[idx]
        self.Ro = self.R = R
        self.dR = dR

        # By default the calculated points are the measured points.  Use
        # oversample() for a more accurate resolution calculations.
        self._set_calc(self.T,self.L)

    @staticmethod
    def alignment_uncertainty(w,I,d=0):
        r"""
        Compute alignment uncertainty.

        **Parameters:**

        *w* : float | degrees
            Rocking curve full width at half max.
        *I* : float | counts
            Rocking curve integrated intensity.
        *d* = 0: float | degrees
            Motor step size

        **Returns:**

        *dtheta* : float | degrees
            uncertainty in alignment angle
        """
        return sqrt(w**2/I + d**2/12.)


    def log10_to_linear(self):
        """
        Convert data from log to linear.

        Older reflectometry reduction code stored reflectivity in log base 10
        format.  Call probe.log10_to_linear() after loading this data to
        convert it to linear for subsequent display and fitting.
        """
        if self.Ro != None:
            self.Ro = 10**self.Ro
            if self.dR != None:
                self.dR = self.Ro * self.dR * log(10)
            self.R = self.Ro

    def _get_data(self):
        raise NotImplementedError
        return self.Ro,self.dR
    def _set_data(self, data):
        # Setting data is dangerous since the Q points may have been
        # reordered to keep them sorted.  The external user may be
        # expecting to reset them from the original data file, which
        # would fail, or from a simulation based on Q, which would
        # succeed.  Storing the reordering vector wouldn't help since
        # it would just reverse the problem and still lead to confusion.
        # Returning self.Q in the original order would allow both cases
        # to work, but it makes simulation more complicated because the
        # resolution calculation wants sorted inputs.  We could just
        # make the resolution calculation sort its inputs, but that
        # increases the cost of computing the reflectivity curve (though
        # that cost may small compared to the cost of sorting an almost
        # sorted list).
        raise NotImplementedError
        if data is not None:
            self.R,self.dR = data
        else:
            self.R = self.dR = None
        # Remember the original so we can resynthesize as needed
        self.Ro = self.R
    data = property(_get_data, _set_data)
    def resynth_data(self):
        """
        Generate new data according to the model R ~ N(Ro,dR).

        The resynthesis step is a precursor to refitting the data, as is
        required for certain types of monte carlo error analysis.
        """
        self.R = self.Ro + numpy.random.randn(*self.Ro.shape)*self.dR

    def restore_data(self):
        """
        Restore the original data.
        """
        self.R = self.Ro

    def simulate_data(self, theory, noise=None):
        """
        Set the data for the probe to R, adding random noise dR.

        If noise is None, then use the uncertainty in the probe.

        As a hack, if noise<0, use the probe uncertainty but don't add
        noise to the data.  Don't depend on this behavior.
        """
        self.Ro = theory[1]+0.

        if numpy.isscalar(noise) and noise < 0:
            # leave the probe uncertainty alone, and don't add noise to the data
            self.R = self.Ro
            return

        if noise is None:
            pass # use existing noise
        else:
            self.dR = 0.01*noise*self.Ro
            self.dR[self.dR==0] = 1e-11

        # Add noise to the theory function
        self.resynth_data()

        # Pretend the noisy theory function is the underlying measured data
        # This allows us to resynthesize later, as needed.
        self.Ro = self.R

    def write_data(self, filename,
                   columns=['Q','R','dR'],
                   header=None):
        """
        Save the data to a file.

        *header* is a string with trailing \\n containing the file header.
        *columns* is a list of column names from Q, dQ, R, dR, L, dL, T, dT.

        The default is to write Q,R,dR data.
        """
        if header == None:
            header = "# %s\n"%' '.join(columns)
        with open(filename,'w') as fid:
            fid.write(header)
            data = numpy.vstack([getattr(self,c) for c in columns])
            numpy.savetxt(fid,data.T)

    def _set_calc(self, T, L):
        Q = TL2Q(T=T, L=L)

        idx = numpy.argsort(Q)
        self.calc_T = T[idx]
        self.calc_L = L[idx]
        self.calc_Qo = Q[idx]

        # Only keep the scattering factors that you need
        self.unique_L = numpy.unique(self.calc_L)
        self._L_idx = numpy.searchsorted(self.unique_L, L)

    @property
    def Q(self):
        if self.theta_offset.value != 0:
            Q = TL2Q(T=self.T+self.theta_offset.value, L=self.L)
            #TODO: this may break the Q order on measurements with varying L
        else:
            Q = self.Qo
        return Q

    @property
    def calc_Q(self):
        if self.theta_offset.value != 0:
            Q = TL2Q(T=self.calc_T+self.theta_offset.value, L=self.calc_L)
            #TODO: this may break the Q order on measurements with varying L
        else:
            Q = self.calc_Qo
        return Q if not self.back_reflectivity else -Q

    def parameters(self):
        return {'intensity':self.intensity,
                'background':self.background,
                'back_absorption':self.back_absorption,
                'theta_offset':self.theta_offset,
                }

    def scattering_factors(self, material, density):
        """
        Returns the scattering factors associated with the material given
        the range of wavelengths/energies used in the probe.
        """
        raise NotImplementedError

    def subsample(self, dQ):
        """
        Select points at most every dQ.

        Use this to speed up computation early in the fitting process.

        This changes the data object, and is not reversible.

        The current algorithm is not picking the "best" Q value, just the
        nearest, so if you have nearby Q points with different quality
        statistics (as happens in overlapped regions from spallation
        source measurements at different angles), then it may choose
        badly.  Simple solutions based on the smallest relative error dR/R
        will be biased toward peaks, and smallest absolute error dR will
        be biased toward valleys.
        """
        # Assumes self contains sorted Qo and associated T,L
        # Note: calc_Qo is already sorted
        Q = numpy.arange(self.Qo[0],self.Qo[-1],dQ)
        idx = numpy.unique(numpy.searchsorted(self.Qo,Q))
        #print len(idx),len(self.Qo)

        self.T, self.dT = self.T[idx],self.dT[idx]
        self.L, self.dL = self.L[idx],self.dL[idx]
        self.Qo, self.dQ = self.Qo[idx],self.dQ[idx]
        if self.R is not None: self.Ro = self.R = self.R[idx]
        if self.dR is not None: self.dR = self.dR[idx]
        self._set_calc(self.T,self.L)

    def resolution_guard(self):
        r"""
        Make sure each measured $Q$ point has at least 5 calculated $Q$
        points contributing to it in the range $[-3\Delta Q,3\Delta Q]$.

        *Not Implemented*
        """
        raise NotImplementedError
        # TODO: implement resolution guard.

    def critical_edge(self,  substrate=None, surface=None,
                      n=51, delta=0.25):
        r"""
        Oversample points near the critical edge.

        The critical edge is defined by the difference in scattering
        potential for the *substrate* and *surface* materials, or the
        reverse if *back_reflectivity* is true.

        *n* is the number of $Q$ points to compute near the critical edge.

        *delta* is the relative uncertainty in the material density,
        which defines the range of values which are calculated.

        The $n$ points $Q_i$ are evenly distributed around the critical
        edge in $Q_c \pm \delta Q_c$ by varying angle $\theta$ for a
        fixed wavelength $< \lambda >$, the average of all wavelengths
        in the probe.

        Specifically:

        .. math::

            Q_c^2 &= 16 \pi (\rho - \rho_\text{incident}) \\
            Q_i &= Q_c - \delta_i Q_c (i - (n-1)/2)
                \qquad \text{for} \; i \in 0 \ldots n-1 \\
            \lambda_i &= < \lambda > \\
            \theta_i &= \sin^{-1}(Q_i \lambda_i / 4 \pi)

        If $Q_c$ is imaginary, then $-|Q_c|$ is used instead, so this
        routine can be used for reflectivity signals which scan from
        back reflectivity to front reflectivity.  For completeness,
        the angle $\theta = 0$ is added as well.
        """
        Srho,Sirho = (0,0) if substrate is None else substrate.sld(self)[:2]
        Vrho,Virho = (0,0) if surface is None else surface.sld(self)[:2]
        drho = Srho-Vrho if not self.back_reflectivity else Vrho-Srho
        Q_c = sign(drho)*sqrt(16*pi*abs(drho)*1e-6)
        Q = numpy.linspace(Q_c*(1 - delta), Q_c*(1+delta), n)
        L = numpy.average(self.L)
        T = QL2T(Q=Q, L=L)
        T = numpy.hstack((self.T,T,0))
        L = numpy.hstack((self.L,[L]*(n+1)))
        #print Q
        self._set_calc(T,L)


    def oversample(self, n=20, seed=1):
        """
        Generate an over-sampling of Q to avoid aliasing effects.

        Oversampling is needed for thick layers, in which the underlying
        reflectivity oscillates so rapidly in Q that a single measurement
        has contributions from multiple Kissig fringes.

        Sampling will be done using a pseudo-random generator so that
        accidental structure in the function does not contribute to the
        aliasing.  The generator will usually be initialized with a fixed
        *seed* so that the point selection will not change from run to run,
        but a *seed* of None will choose a different set of points each time
        oversample is called.

        The value *n* is the number of points that should contribute to
        each Q value when computing the resolution.   These will be
        distributed about the nominal measurement value, but varying in
        both angle and energy according to the resolution function.  This
        will yield more points near the measurement and fewer farther away.
        The measurement point itself will not be used to avoid accidental
        bias from uniform Q steps.  Depending on the problem, a value of
        *n* between 20 and 100 should lead to stable values for the convolved
        reflectivity.
        """
        if n<=5:
            raise ValueError("Oversampling with n<=5 is not useful")

        rng = numpy.random.RandomState(seed=seed)
        T = rng.normal(self.T[:,None],self.dT[:,None],size=(len(self.dT),n-1))
        L = rng.normal(self.L[:,None],self.dL[:,None],size=(len(self.dL),n-1))
        T = numpy.hstack((self.T,T.flatten()))
        L = numpy.hstack((self.L,L.flatten()))
        self._set_calc(T,L)

    def _apply_resolution(self, Qin, Rin):
        """
        Apply the instrument resolution function
        """
        R = convolve(Qin, Rin, self.Q, self.dQ)
        return self.Q, R

    def apply_beam(self, calc_Q, calc_R, resolution=True):
        """
        Apply factors such as beam intensity, background, backabsorption,
        resolution to the data.
        """
        # Note: in-place vector operations are not notably faster.

        # Handle absorption through the substrate, which occurs when Q<0
        # (condition)*C is C when condition is True or 0 when False,
        # (condition)*(C-1)+1 is C when condition is True or 1 when False.
        back = (calc_Q<0)*(self.back_absorption.value-1)+1
        calc_R = calc_R * back

        # For back reflectivity, reverse the sign of Q after computing
        if self.back_reflectivity:
            calc_Q = -calc_Q
        if calc_Q[-1] < calc_Q[0]:
            calc_Q, calc_R = [v[::-1] for v in (calc_Q, calc_R)]
        if resolution:
            Q,R = self._apply_resolution(calc_Q, calc_R)
        else:
            # Given that the target Q points should be in the set of
            # calculated Q values, interp will give us the
            # values of Q at the appropriate R, even if there are
            # guard values, or otherwise some mixture of calculated
            # Q values.  The cost of doing so is going to be n log n
            # in the size of Q, which is a bit pricey, but let's see
            # if it is a problem before optimizing.
            Q,R = self.Q, numpy.interp(self.Q, calc_Q, calc_R)
        R = self.intensity.value*R + self.background.value
        return Q,R

    def fresnel(self, substrate=None, surface=None):
        """
        Returns a Fresnel reflectivity calculator given the surface and
        and substrate.  The calculated reflectivity includes The Fresnel reflectivity for the probe reflecting from a block of
        material with the given substrate.

        Returns F = R(probe.Q), where R is magnitude squared reflectivity.
        """
        # Doesn't use ProbeCache, but this routine is not time critical
        Srho,Sirho = (0,0) if substrate is None else substrate.sld(self)[:2]
        Vrho,Virho = (0,0) if surface is None else surface.sld(self)[:2]
        if self.back_reflectivity:
            Srho,Vrho = Vrho,Srho
            Sirho,Virho = Virho,Sirho
        if Srho == Vrho: Srho = Vrho + 1
        #I = numpy.ones_like(self.Q)
        I = 1
        calculator = fresnel.Fresnel(rho=Srho*I, irho=Sirho*I,
                                     Vrho=Vrho*I, Virho=Virho*I)
        return calculator

    def save(self, filename, theory, substrate=None, surface=None):
        """
        Save the data and theory to a file.
        """
        fresnel = self.fresnel(substrate, surface)
        _, R = theory
        fid = open(filename,"w")
        fid.write("# intensity: %.15g\n# background: %.15g\n"
                  % (self.intensity.value, self.background.value))
        if getattr(self,'R',None) != None:
            A = numpy.array((self.Q,self.dQ,self.R,self.dR, R, fresnel(self.Q)))
            fid.write("# %17s %20s %20s %20s %20s %20s\n"
                      %("Q (1/A)","dQ (1/A)", "R", "dR", "theory", "fresnel"))
        else:
            A = numpy.array((self.Q,self.dQ,R,fresnel(self.Q)))
            fid.write("# %17s %20s %20s %20s\n"
                      %("Q (1/A)","dQ (1/A)", "theory", "fresnel"))
        #print "A",self.Q.shape,A.shape
        numpy.savetxt(fid, A.T, fmt="%20.15g")
        fid.close()

    def plot(self, view=None, **kwargs):
        """
        Plot theory against data.

        Need substrate/surface for Fresnel-normalized reflectivity
        """

        view = view if view is not None else self.view

        if view == 'linear':
            self.plot_linear(**kwargs)
        elif view == 'log':
            self.plot_log(**kwargs)
        elif view == 'fresnel':
            self.plot_fresnel(**kwargs)
        elif view == 'logfresnel':
            self.plot_logfresnel(**kwargs)
        elif view == 'q4':
            self.plot_Q4(**kwargs)
        elif view == 'resolution':
            self.plot_resolution(**kwargs)
        elif view.startswith('resid'):
            self.plot_residuals(**kwargs)
        elif view == 'fft':
            self.plot_fft(**kwargs)
        elif view == 'SA': # SA uses default plot
            self.plot(view=None, **kwargs)
        else:
            raise TypeError("incorrect reflectivity view '%s'"%view)


    def plot_resolution(self, suffix='', label=None, **kwargs):
        import pylab
        pylab.plot(self.Q, self.dQ,
                   label=self.label(prefix=label,suffix=suffix))
        pylab.xlabel(r'Q ($\AA^{-1}$)')
        pylab.ylabel(r'Q resolution ($1-\sigma \AA^{-1}$)')
        pylab.title('Measurement resolution')


    def plot_linear(self, **kwargs):
        """
        Plot the data associated with probe.
        """
        import pylab
        self._plot_pair(ylabel='Reflectivity', **kwargs)
        pylab.yscale('linear')

    def plot_log(self, **kwargs):
        """
        Plot the data associated with probe.
        """
        import pylab
        self._plot_pair(ylabel='Reflectivity', **kwargs)
        pylab.yscale('log')

    def plot_logfresnel(self, *args, **kw):
        """
        Plot the log Fresnel-normalized reflectivity associated with the probe.
        """
        import pylab
        self.plot_fresnel(*args, **kw)
        pylab.yscale('log')

    def plot_fresnel(self, substrate=None, surface=None, **kwargs):
        r"""
        Plot the Fresnel-normalized reflectivity associated with the probe.

        Note that the Fresnel reflectivity has the intensity and background 
        applied before normalizing so that hydrogenated samples display more 
        cleanly.  The formula to reproduce the graph is:

        .. math::

             R' = R / (F(Q) I + B)
             \Delta R' = \Delta R / (F(Q) I + B)

        where $I$ is the intensity and $B$ is the background.
        """
        import pylab
        if substrate is None and surface is None:
            raise TypeError("Fresnel-normalized reflectivity needs substrate or surface")
        F = self.fresnel(substrate=substrate,surface=surface)
        scale = lambda Q,dQ,R,dR: (
            Q,dQ,
            R/(F(Q)*self.intensity.value+self.background.value),
            dR/(F(Q)*self.intensity.value+self.background.value))
        if substrate is None:
            name = "air:%s"%(surface.name)
        elif surface is None or isinstance(surface,Vacuum):
            name = substrate.name
        else:
            name = "%s:%s"%(substrate.name, surface.name)
        self._plot_pair(correct=scale, ylabel='R/(R(%s)'%(name), **kwargs)

    def plot_Q4(self, **kwargs):
        r"""
        Plot the Q**4 reflectivity associated with the probe.  

        Note that Q**4 reflectivity has the intensity and background applied
        so that hydrogenated samples display more cleanly.  The formula
        to reproduce the graph is:

        .. math::

             R' = R / ( (100*Q)^{-4} I + B)
             \Delta R' = \Delta R / ( (100*Q)^{-4} I + B )

        where $B$ is the background.
        """
        import pylab
        scale = lambda Q,dQ,R,dR: (
            Q,dQ,
            #R/numpy.maximum(1e-8*Q**-4,self.background.value),
            #dR/numpy.maximum(1e-8*Q**-4,self.background.value))
            R/(1e-8*Q**-4*self.intensity.value + self.background.value),
            dR/(1e-8*Q**-4*self.intensity.value + self.background.value))
        #Q4[Q4==0] = 1
        self._plot_pair(correct=scale, ylabel='R (100 Q)^4', **kwargs)

    def _plot_pair(self, theory=None, correct=lambda Q,dQ,R,dR: (Q,dQ,R,dR),
                   ylabel="", suffix="", label=None,
                   plot_shift=None, **kw):
        import pylab
        c = coordinated_colors()
        isheld = pylab.ishold()
        plot_shift = plot_shift if plot_shift is not None else Probe.plot_shift
        trans = auto_shift(plot_shift)
        if hasattr(self,'R') and self.R is not None:
            Q, dQ, R, dR = correct(self.Q, self.dQ, self.R, self.dR)
            pylab.errorbar(Q, R, yerr=dR, xerr=dQ, capsize=0,
                           fmt='.', color=c['light'], transform=trans,
                           label=self.label(prefix=label,
                                            gloss='data',
                                            suffix=suffix))
            pylab.hold(True)
        if theory is not None:
            Q,R = theory
            Q, dQ, R, dR = correct(Q, 0, R, 0)
            pylab.plot(Q, R,
                       color=c['dark'], transform=trans,
                       label=self.label(prefix=label,
                                        gloss='theory',
                                        suffix=suffix))
        pylab.hold(isheld)
        pylab.xlabel('Q (inv Angstroms)')
        pylab.ylabel(ylabel)
        #pylab.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        h = pylab.legend(fancybox=True, numpoints=1)
        h.get_frame().set_alpha(0.5)

    def plot_residuals(self, theory=None, suffix='', label=None,
                       plot_shift=None, **kwargs):
        import pylab
        plot_shift = plot_shift if plot_shift is not None else Probe.residuals_shift
        trans = auto_shift(plot_shift)
        if theory is not None and self.R is not None:
            c = coordinated_colors()
            Q,R = theory
            residual = (R - self.R)/self.dR
            pylab.plot(self.Q, residual,
                       '.', color=c['light'],
                       transform=trans,
                       label=self.label(prefix=label, suffix=suffix))
        pylab.axhline(1, color='black', ls='--',lw=1)
        pylab.axhline(0, color='black', lw=1)
        pylab.axhline(-1, color='black', ls='--',lw=1)
        pylab.xlabel('Q (inv A)')
        pylab.ylabel('(theory-data)/error')
        pylab.legend(numpoints=1)

    def plot_fft(self, theory=None, suffix='', label=None,
                 substrate=None, surface=None, **kwargs):
        """
        FFT analysis of reflectivity signal.
        """
        raise NotImplementedError
        import pylab
        c = coordinated_colors()
        isheld = pylab.ishold()
        if substrate is None and surface is None:
            raise TypeError("FFT reflectivity needs substrate or surface")
        F = self.fresnel(substrate=substrate,surface=surface)
        #Qc = sqrt(16*pi*substrate)
        Qc = 0
        Qmax = max(self.Q)
        T = numpy.linspace(Qc,Qmax,len(self.Q))
        z = numpy.linspace(0,2*pi/Qmax, len(self.Q)//2)
        if hasattr(self,'R'):
            signal = numpy.interp(T, self.Q, self.R/F(self.Q))
            A = abs(numpy.fft.fft(signal - numpy.average(signal)))
            A = A[:len(A)//2]
            pylab.plot(z, A, '.-', color=c['light'],
                       label=self.label(prefix=label,
                                        gloss='data',
                                        suffix=suffix))
            pylab.hold(True)
        if theory is not None:
            Q,R = theory
            signal = numpy.interp(T, Q, R/F(Q))
            A = abs(numpy.fft.fft(signal-numpy.average(signal)))
            A = A[:len(A)//2]
            pylab.plot(z, A, '-', color=c['dark'],
                       label=self.label(prefix=label,
                                        gloss='theory',
                                        suffix=suffix))
        pylab.hold(isheld)
        pylab.xlabel('w (A)')
        if substrate is None:
            name = "air:%s"%(surface.name)
        elif surface is None or isinstance(surface,Vacuum):
            name = substrate.name
        else:
            name = "%s:%s"%(substrate.name, surface.name)
        pylab.ylabel('|FFT(R/R(%s))|'%(name))

    def label(self, prefix=None, gloss="", suffix=""):
        if not prefix:
            prefix = self.name
        if prefix:
            return " ".join((prefix+suffix,gloss)) if gloss else prefix
        else:
            return suffix+" "+gloss if gloss else None

class XrayProbe(Probe):
    """
    X-Ray probe.

    By providing a scattering factor calculator for X-ray scattering, model
    components can be defined by mass density and chemical composition.
    """
    radiation = "xray"
    def scattering_factors(self, material, density):
        # doc string is inherited from parent (see below)
        # Note: the real density is calculated as a scale factor applied to
        # the returned sld as computed assuming density=1
        rho, irho = xsf.xray_sld(material,
                                 wavelength = self.unique_L,
                                 density = density)
        # TODO: support wavelength dependent systems
        return rho[0], irho[0], 0
        return rho[self._L_idx], irho[self._L_idx], 0
    scattering_factors.__doc__ = Probe.scattering_factors.__doc__

class NeutronProbe(Probe):
    """
    Neutron probe.

    By providing a scattering factor calculator for X-ray scattering, model
    components can be defined by mass density and chemical composition.
    """
    radiation = "neutron"
    def scattering_factors(self, material, density):
        # doc string is inherited from parent (see below)
        # Note: the real density is calculated as a scale factor applied to
        # the returned sld as computed assuming density=1
        rho, irho, rho_incoh = nsf.neutron_sld(material,
                                               wavelength = self.unique_L,
                                               density = density)
        # TODO: support wavelength dependent systems
        return rho, irho[0], rho_incoh
        return rho, irho[self._L_idx], rho_incoh
    scattering_factors.__doc__ = Probe.scattering_factors.__doc__


class ProbeSet(Probe):
    def __init__(self, probes, name=None):
        self.probes = list(probes)
        self.R = numpy.hstack(p.R for p in self.probes)
        self.dR = numpy.hstack(p.dR for p in self.probes)
        self.dQ = numpy.hstack(p.dQ for p in self.probes)
        self.name = name if name is not None else self.probes[0].name

        back_refls = [f.back_reflectivity for f in self.probes]
        if all(back_refls) or not any(back_refls):
            self.back_reflectivity = back_refls[0]
        else:
            raise ValueError("Cannot mix front and back reflectivity measurements")

    def parameters(self):
        return [p.parameters() for p in self.probes]
    parameters.__doc__ = Probe.parameters.__doc__

    def resynth_data(self):
        for p in self.probes: p.resynth_data()
        self.R = numpy.hstack(p.R for p in self.probes)
    resynth_data.__doc__ = Probe.resynth_data.__doc__

    def restore_data(self):
        for p in self.probes: p.restore_data()
        self.R = numpy.hstack(p.R for p in self.probes)
    restore_data.__doc__ = Probe.restore_data.__doc__

    def simulate_data(self, theory, noise=2):
        Q,R = theory
        offset = 0
        for p in self.probes:
            n = len(p.Q)
            p.simulate_data(theory=(Q[offset:offset+n],R[offset:offset+n]),
                            noise=noise)
            offset += n
    simulate_data.__doc__ = Probe.simulate_data.__doc__

    @property
    def Q(self):
        return numpy.hstack(p.Q for p in self.probes)
    @property
    def calc_Q(self):
        return numpy.unique(numpy.hstack(p.calc_Q for p in self.probes))
    @property
    def unique_L(self):
        return numpy.unique(numpy.hstack(p.unique_L for p in self.probes))
    def oversample(self, **kw):
        for p in self.probes: p.oversample(**kw)
    oversample.__doc__ = Probe.oversample.__doc__
    def scattering_factors(self, material, density):
        # TODO: support wavelength dependent systems
        return self.probes[0].scattering_factors(material, density)
        result = [p.scattering_factors(material, density) for p in self.probes]
        return [numpy.hstack(v) for v in zip(*result)]
    scattering_factors.__doc__ = Probe.scattering_factors.__doc__
    def apply_beam(self, calc_Q, calc_R, resolution=True, **kw):
        result = [p.apply_beam(calc_Q, calc_R, **kw) for p in self.probes]
        Q,R = [numpy.hstack(v) for v in zip(*result)]
        return Q,R
    def fresnel(self, *args, **kw):
        return self.probes[0].fresnel(*args, **kw)
    fresnel.__doc__ = Probe.fresnel.__doc__
    def save(self, filename, theory, substrate=None, surface=None):
        for i,(p,th) in enumerate(self.parts(theory=theory)):
            p.save(filename+str(i+1), th, substrate=substrate, surface=surface)
    save.__doc__ = Probe.save.__doc__
    def plot(self, theory=None, **kw):
        import pylab
        ishold = pylab.ishold()
        for p,th in self.parts(theory):
            p.plot(theory=th, **kw)
            pylab.hold(True)
        pylab.hold(ishold)
    plot.__doc__ = Probe.plot.__doc__
    def plot_resolution(self, **kw):
        for p in self.probes: p.plot_resolution(**kw)
    plot_resolution.__doc__ = Probe.plot_resolution.__doc__
    def plot_linear(self, theory=None, **kw):
        for p,th in self.parts(theory): p.plot_linear(theory=th, **kw)
    plot_linear.__doc__ = Probe.plot_linear.__doc__
    def plot_log(self, theory=None, **kw):
        for p,th in self.parts(theory): p.plot_log(theory=th, **kw)
    plot_log.__doc__ = Probe.plot_log.__doc__
    def plot_fresnel(self, theory=None, **kw):
        for p,th in self.parts(theory): p.plot_fresnel(theory=th, **kw)
    plot_fresnel.__doc__ = Probe.plot_fresnel.__doc__
    def plot_logfresnel(self, theory=None, **kw):
        for p,th in self.parts(theory): p.plot_logfresnel(theory=th, **kw)
    plot_logfresnel.__doc__ = Probe.plot_logfresnel.__doc__
    def plot_Q4(self, theory=None, **kw):
        for p,th in self.parts(theory): p.plot_Q4(theory=th, **kw)
    plot_Q4.__doc__ = Probe.plot_Q4.__doc__
    def plot_residuals(self, theory=None, **kw):
        for p,th in self.parts(theory): p.plot_residuals(theory=th, **kw)
    plot_residuals.__doc__ = Probe.plot_residuals.__doc__
    def parts(self, theory):
        if theory == None:
            for p in self.probes:
                yield p,None
        else:
            offset = 0
            Q,R = theory
            for p in self.probes:
                n = len(p.Q)
                yield p,(Q[offset:offset+n],R[offset:offset+n])
                offset += n

    def shared_beam(self, intensity=1, background=0,
                    back_absorption=1, theta_offset=0):
        """
        Share beam parameters across all segments.

        New parameters are created for *intensity*, *background*,
        *theta_offset* and *back_absorption* and assigned to the all
        segments.  These can be replaced in an individual segment if
        that parameter is independent.
        """
        intensity = Parameter.default(intensity,name="intensity")
        background = Parameter.default(background,name="background",
                                       limits=[0,inf])
        back_absorption = Parameter.default(back_absorption,
                                            name="back_absorption",
                                            limits=[0,1])
        theta_offset = Parameter.default(theta_offset,name="theta_offset")
        for p in self.probes:
            p.intensity = intensity
            p.background = background
            p.back_absorption = back_absorption
            p.theta_offset = theta_offset

    def stitch(self, same_Q=0.001, same_dQ=0.001):
        r"""
        Stitch together multiple datasets into a single dataset.

        Points within *tol* of each other and with the same resolution
        are combined by interpolating them to a common $Q$ value then averaged
        using Gaussian error propagation.

        :Returns: probe | Probe
            Combined data set.

        :Algorithm:

        To interpolate a set of points to a common value, first find the
        common $Q$ value:

        .. math::

            \hat Q = \sum{Q_k} / n

        Then for each dataset $k$, find the interval $[i,i+1]$ containing the
        value $Q$, and use it to compute interpolated value for $R$:

        .. math::

            w &= (\hat Q - Q_i)/(Q_{i+1} - Q_i) \\
            \hat R &= w R_{i+1} + (1-w) R_{i+1} \\
            \hat \sigma_{R} &=
                \sqrt{ w^2 \sigma^2_{R_i} + (1-w)^2 \sigma^2_{R_{i+1}} } / n

        Average the resulting $R$ using Gaussian error propagation:

        .. math::

            \hat R &= \sum{\hat R_k}/n \\
            \hat \sigma_R &= \sqrt{\sum \hat \sigma_{R_k}^2}/n

        """
        Q,dQ,R,dR = stitch(self.probes)
        Po = self.probes[0]
        return QProbe(Q,dQ,data=(R,dR),
                      intensity=Po.intensity,
                      background=Po.background,
                      back_absorption=Po.back_absorption,
                      back_reflectivity=Po.back_reflectivity)

class QProbe(Probe):
    """
    A pure Q,R probe

    This probe with no possibility of tricks such as looking up the
    scattering length density based on wavelength, or adjusting for
    alignment errors.
    """
    def __init__(self, Q, dQ, data=None, name=None,
                 intensity=1, background=0, back_absorption=1,
                 back_reflectivity=False):
        qualifier = " "+name if name is not None else ""
        self.intensity = Parameter.default(intensity,name="intensity"+qualifier)
        self.background = Parameter.default(background,name="background"+qualifier,
                                            limits=[0,inf])
        self.back_absorption = Parameter.default(back_absorption,
                                                 name="back_absorption"+qualifier,
                                                 limits=[0,1])
        self.theta_offset = Constant(0,name="theta_offset"+qualifier)

        self.back_reflectivity = back_reflectivity


        self.Qo, self.dQ = Q, dQ
        if data is not None:
            R,dR = data
        else:
            R,dR = None,None

        self.Qo, self.dQ = Q,dQ
        self.Ro = self.R = R
        self.dR = dR
        self.unique_L = None
        self.calc_Qo = self.Qo
        self.name = name


def measurement_union(xs):
    """
    Determine the unique (T,dT,L,dL) across all datasets.
    """

    # First gather a set of unique tuples in angle and wavelength
    TL = set()
    for x in xs:
        if x is not None:
            TL = TL | set(zip(x.T,x.dT,x.L,x.dL))
    T,dT,L,dL = zip(*[item for item in TL])
    T,dT,L,dL = [numpy.asarray(v) for v in (T,dT,L,dL)]

    # Convert to Q,dQ
    Q = TL2Q(T,L)
    dQ = dTdL2dQ(T,dT,L,dL)

    # Sort by Q
    idx = numpy.argsort(Q)
    T,dT,L,dL,Q,dQ = T[idx],dT[idx],L[idx],dL[idx],Q[idx],dQ[idx]
    if abs(Q[1:] - Q[:-1]).any() < 1e-14:
        raise ValueError("Q is not unique")
    return T, dT, L, dL, Q, dQ

def Qmeasurement_union(xs):
    """
    Determine the unique Q,dQ across all datasets.
    """
    Qset = set()
    for x in xs:
        if x is not None:
            Qset = Qset | set(zip(x.Q,x.dQ))
    Q,dQ = [numpy.array(sorted(v))
            for v in zip(*[w for w in Qset])]
    idx = numpy.argsort(Q)
    Q,dQ = Q[idx],dQ[idx]
    if abs(Q[1:] - Q[:-1]).any() < 1e-14:
        raise ValueError("Q is not unique")
    return Q, dQ

class PolarizedNeutronProbe(object):
    """
    Polarized neutron probe

    *xs* (4 x NeutronProbe) is a sequence pp, pm, mp and mm.
    *Aguide* (degrees) is the angle of the guide field
    """
    view = None  # Default to Probe.view so only need to change in one place
    substrate = surface = None
    polarized = True
    def __init__(self, xs=None, name=None, Aguide=270):
        self.xs = xs

        self.name = name if name is not None else self.xs[0].name
        self.T, self.dT, self.L, self.dL, self.Q, self.dQ \
            = measurement_union(xs)
        self._set_calc(self.T, self.L)
        self._check()
        self.Aguide = Aguide

    @property
    def pp(self): return self.xs[3]
    @property
    def pm(self): return self.xs[1]
    @property
    def mp(self): return self.xs[1]
    @property
    def mm(self): return self.xs[0]

    def parameters(self):
        mm,mp,pm,pp = [(xsi.parameters() if xsi else None)
                       for xsi in self.xs]
        return {'pp':pp, 'pm':pm, 'mp':mp, 'mm':mm}

    def resynth_data(self):
        for p in self.xs:
            if p is not None: p.resynth_data()
    resynth_data.__doc__ = Probe.resynth_data.__doc__

    def restore_data(self):
        for p in self.xs:
            if p is not None: p.restore_data()
    restore_data.__doc__ = Probe.restore_data.__doc__

    def simulate_data(self, theory, noise=2):
        for p,th in zip(self.xs,theory):
            if p is not None: p.simulate_data(theory=th, noise=noise)
    simulate_data.__doc__ = Probe.simulate_data.__doc__

    def _check(self):
        back_refls = [f.back_reflectivity for f in self.xs if f is not None]
        if all(back_refls) or not any(back_refls):
            self.back_reflectivity = back_refls[0]
        else:
            raise ValueError("Cannot mix front and back reflectivity measurements")

    def shared_beam(self, intensity=1, background=0,
                    back_absorption=1, theta_offset=0):
        """
        Share beam parameters across all four cross sections.

        New parameters are created for *intensity*, *background*,
        *theta_offset* and *back_absorption* and assigned to the all
        cross sections.  These can be replaced in an individual
        cross section if for some reason one of the parameters is
        independent.
        """
        intensity = Parameter.default(intensity,name="intensity")
        background = Parameter.default(background,name="background",
                                       limits=[0,inf])
        back_absorption = Parameter.default(back_absorption,
                                            name="back_absorption",
                                            limits=[0,1])
        theta_offset = Parameter.default(theta_offset,name="theta_offset")
        for x in self.xs:
            if x is not None:
                x.intensity = intensity
                x.background = background
                x.back_absorption = back_absorption
                x.theta_offset = theta_offset

    def oversample(self, n=6, seed=1):
        # doc string is inherited from parent (see below)
        rng = numpy.random.RandomState(seed=seed)
        T = rng.normal(self.T[:,None],self.dT[:,None],size=(len(self.dT),n))
        L = rng.normal(self.L[:,None],self.dL[:,None],size=(len(self.dL),n))
        T = T.flatten()
        L = L.flatten()
        self._set_calc(T,L)
    oversample.__doc__ = Probe.oversample.__doc__

    @property
    def calc_Q(self):
        # All calc_Q should be the same, so return the first one
        for x in self.xs:
            if x is not None:
                return x.calc_Q
        raise RuntimeError("No polarization cross sections")

    def _set_calc(self, T, L):
        """
        Propagate setting of calc_Q to the individual cross sections.
        """
        # Set all xs to have the same calc_Q
        for x in self.xs:
            if x is not None:
                x._set_calc(T,L)
        # Only keep the scattering factors that you need
        self.unique_L = numpy.unique(L)
        self._L_idx = numpy.searchsorted(self.unique_L, L)

    def apply_beam(self, Q, R, resolution=True):
        """
        Apply factors such as beam intensity, background, backabsorption,
        and footprint to the data.
        """
        return [(xs.apply_beam(Q,Ri,resolution) if xs else None)
                for xs,Ri in zip(self.xs,R)]

    def fresnel(self, *args, **kw):
        return self.pp.fresnel(*args, **kw)
    fresnel.__doc__ = Probe.fresnel.__doc__

    def scattering_factors(self, material, density):
        # doc string is inherited from parent (see below)
        rho, irho, rho_incoh = nsf.neutron_sld(material,
                                               wavelength=self.unique_L,
                                               density=density)
        # TODO: support wavelength dependent systems
        return rho, irho[0], rho_incoh
        return rho, irho[self._L_idx], rho_incoh
    scattering_factors.__doc__ = Probe.scattering_factors.__doc__

    def select_corresponding(self, theory):
        """
        Select theory points corresponding to the measured data.

        Since we have evaluated theory at every Q, it is safe to interpolate
        measured Q into theory, since it will land on a node,
        not in an interval.
        """

        Qth,Rth = theory
        return [None if x_data is None
                else (x_data.Q, numpy.interp(x_data.Q,Qth,x_th))
                for x_data,x_th in zip(self.xs,Rth)]


    def save(self, filename, theory, substrate=None, surface=None):
        for xsi,xsi_th,suffix in zip(self.xs, theory, ('A','B','C','D')):
            if xsi is not None:
                xsi.save(filename+suffix, xsi_th,
                         substrate=substrate, surface=surface)
    save.__doc__ = Probe.save.__doc__

    def plot(self, view=None, **kwargs):
        """
        Plot theory against data.

        Need substrate/surface for Fresnel-normalized reflectivity
        """
        view = view if view is not None else self.view

        if view is None: view = Probe.view  # Default to Probe.view

        if view == 'linear':
            self.plot_linear(**kwargs)
        elif view == 'log':
            self.plot_log(**kwargs)
        elif view == 'fresnel':
            self.plot_fresnel(**kwargs)
        elif view == 'logfresnel':
            self.plot_logfresnel(**kwargs)
        elif view == 'q4':
            self.plot_Q4(**kwargs)
        elif view == 'residuals':
            self.plot_residuals(**kwargs)
        elif view == 'SA':
            self.plot_SA(**kwargs)
        elif view == 'resolution':
            self.plot_resolution(**kwargs)
        else:
            raise TypeError("incorrect reflectivity view '%s'"%view)

    def plot_resolution(self, **kwargs):
        self._xs_plot('plot_resolution', **kwargs)
    def plot_linear(self, **kwargs):
        self._xs_plot('plot_linear', **kwargs)
    def plot_log(self, **kwargs):
        self._xs_plot('plot_log', **kwargs)
    def plot_fresnel(self, **kwargs):
        self._xs_plot('plot_fresnel', **kwargs)
    def plot_logfresnel(self, **kwargs):
        self._xs_plot('plot_logfresnel', **kwargs)
    def plot_Q4(self, **kwargs):
        self._xs_plot('plot_Q4', **kwargs)
    def plot_residuals(self, **kwargs):
        self._xs_plot('plot_residuals', **kwargs)
    def plot_SA(self, theory=None, label=None, plot_shift=None,
                **kwargs):
        import pylab
        if self.pp is None or self.mm is None:
            raise TypeError("cannot form spin asymmetry plot without ++ and --")

        plot_shift = plot_shift if plot_shift is not None else Probe.plot_shift
        trans = auto_shift(plot_shift)
        isheld = pylab.ishold()
        pp,mm = self.pp,self.mm
        c = coordinated_colors()
        if hasattr(pp,'R'):
            Q,SA,dSA = spin_asymmetry(pp.Q,pp.R,pp.dR,mm.Q,mm.R,mm.dR)
            if dSA is not None:
                pylab.errorbar(Q, SA, yerr=dSA, xerr=pp.dQ, fmt='.', capsize=0,
                               label=pp.label(prefix=label, gloss='data'),
                               transform=trans,
                               color=c['light'])
            else:
                pylab.plot(Q,SA,'.',
                           label=pp.label(prefix=label,gloss='data'),
                               transform=trans,
                           color=c['light'])
            pylab.hold(True)
        if theory is not None:
            mm,mp,pm,pp = theory
            Q,SA,_ = spin_asymmetry(pp[0],pp[1],None,mm[0],mm[1],None)
            pylab.plot(Q, SA,
                       label=self.pp.label(prefix=label,gloss='theory'),
                       transform=trans,
                       color=c['dark'])
        pylab.hold(isheld)
        pylab.xlabel(r'Q (\AA^{-1})')
        pylab.ylabel(r'spin asymmetry $(R^{++} -\, R^{--}) / (R^{++} +\, R^{--})$')
        pylab.legend(numpoints=1)

    def _xs_plot(self, plotter, theory=None, **kwargs):
        import pylab
        # Plot available cross sections
        isheld = pylab.ishold()
        if theory is None:
            theory = (None,None,None,None)
        for x_data,x_th,suffix in zip(self.xs, theory,
                                      ('$^{--}$','$^{-+}$','$^{+-}$','$^{++}$')):
            if x_data is not None:
                fn = getattr(x_data, plotter)
                fn(theory=x_th, suffix=suffix, **kwargs)
                pylab.hold(True)
        if not isheld: pylab.hold(False)

def spin_asymmetry(Qp,Rp,dRp,Qm,Rm,dRm):
    r"""
    Compute spin asymmetry for R++, R--.

    **Parameters:**

    *Qp*, *Rp*, *dRp* : vector
        Measured ++ cross section and uncertainty.
    *Qm*, *Rm*, *dRm* : vector
        Measured -- cross section and uncertainty.

    If *dRp*, *dRm* are None then the returned uncertainty will also be None.

    **Returns:**

    *Q*, *SA*, *dSA* : vector
        Computed spin asymmetry and uncertainty.

    **Algorithm:**

    Spin asymmetry, $S_A$, is:

    .. math::

        S_A = \frac{R_{++} - R_{--}}{R_{++} + R_{--}}

    Uncertainty $\Delta S_A$ follows from propagation of error:

    .. math::

        \Delta S_A^2 = \frac{4(R_{++}^2\Delta R_{--}^2+R_{--}^2\Delta R_{++})}
                            {(R_{++} + R_{--})^4}

    """
    Rm = numpy.interp(Qp,Qm,Rm)
    v = (Rp-Rm)/(Rp+Rm)
    if dRp is not None:
        dRm = numpy.interp(Qp,Qm,dRm)
        dvsq = 4 * ((Rp*dRm)**2 + (Rm*dRp)**2) / (Rp+Rm)**4
        dvsq[dvsq<0] = 0
        return Qp, v, sqrt(dvsq)
    else:
        return Qp, v, None


class PolarizedNeutronQProbe(PolarizedNeutronProbe):
    polarized = True
    def __init__(self, xs=None, name=None, Aguide=270):
        self.xs = xs
        self.Q, self.dQ = Qmeasurement_union(xs)
        self._check()
        self.name = name if name is not None else xs[0].name
        self.Aguide = Aguide




"""
Residual correlation test.

INCOMPLETE UNUSED UNTESTED CODE

Returns the probability of seeing the given structure in the residuals.

The residuals are assumed to be e ~ N(0,1).  Gross deviations from this
form will already be accounted for in the chisq statistic, and can be
ignored here.  We do this by translating the residual into a z-scored
residual, which should be in

Protects against an incorrect estimation of the uncertainty in the data,
which is anyway already covered by the chisq cost function.
"""

from scipy import stats

def residual_nllf(v):
    # Normalize the scores so that we are assuming
    z = stats.zs(v)



def plot(self):
    pylab.subplot(313)
    Q,R = self.reflectivity()
    self.probe.plot_transform(theory=(Q,R),
                    substrate=self.sample[0].material,
                    surround=self.sample[-1].material)

# Experiments with different sorts of
def plot_transform(self, theory=None, substrate=None, surround=None):
    """
    Plot the Fresnel-normalized reflectivity associated with the probe.
    """
    import pylab
    if substrate is None and surround is None:
        raise TypeError("Fresnel-normalized reflectivity needs substrate or surround")
    F = self.fresnel(substrate=substrate,surround=surround)
    print("substrate %s"%substrate)
    #Qc = sqrt(16*pi*substrate)
    Qc = 0
    T = numpy.linspace(Qc,max(self.Q),len(self.Q))
    if hasattr(self,'R'):
        A = abs(numpy.fft.fft(numpy.interp(T,self.Q,self.R/F)))
        pylab.plot(T,A, '.')
    if theory is not None:
        Q,R = theory
        A = abs(numpy.fft.fft(numpy.interp(T,self.Q,self.R/F)))
        pylab.plot(T,A, hold=True)
    pylab.xlabel('z (Angstroms)')
    if substrate is None:
        name = "air:%s"%(surround.name)
    elif surround is None or isinstance(surround,Vacuum):
        name = substrate.name
    else:
        name = "%s:%s"%(substrate.name, surround.name)
    pylab.ylabel('FFT(R/R(%s))'%(name))

def plot_deriv(self, theory=None):
    import pylab
    scale = 1e8*self.Q**4
    #scale=1
    if hasattr(self, 'R'):
        d = deriv(self.Q, self.R, self.dR)
        pylab.plot(self.Q, d*scale, '.')
    if theory is not None:
        Q,R = theory
        dth = deriv(Q,R)
        pylab.plot(Q,dth*scale,hold=True)
    #pylab.plot(Q,(sign(d)-sign(dth))/2)
    pylab.xlabel('Q (inv Angstroms)')
    pylab.ylabel('R\' (100 Q)^4')

def deriv(Q, R, dR=None, width=5, degree=2):
    from bumps.wsolve import wpolyfit
    d = numpy.empty_like(Q)
    if dR is None: dR = numpy.ones_like(Q)
    k = (width-1)/2
    p = wpolyfit(Q[:width],R[:width],dy=dR[:width],degree=degree)
    d[:k+1]=p.der(Q[:k+1])
    for i in range(k+1,len(Q)-k-1):
        idx = slice(i-k,i+k+1)
        p = wpolyfit(Q[idx],R[idx],dy=dR[idx],degree=degree)
        d[i]=p.der(Q[i])
    p = wpolyfit(Q[-width:],R[-width:],dy=dR[-width:],degree=degree)
    d[-k-1:] = p.der(Q[-k-1:])
    return d



# -*- coding: utf-8 -*-
# This program is public domain
# Authors Paul Kienzle, Richard Sheridan
"""
Layer models for polymer systems.

Analytic Self-consistent Field (SCF) Brush profile [#Zhulina]_ [#Karim]_

Analytical Self-consistent Field (SCF) Mushroom Profile [#Adamuţi-Trache]_

Numerical Self-consistent Field (SCF) End-Tethered Polymer Profile [#Cosgrove]_ [#deVos]_ [#Sheridan]_


.. [#Zhulina] Zhulina, EB; Borisov, OV; Pryamitsyn, VA; Birshtein, TM (1991)
   "Coil-Globule Type Transitions in Polymers. 1. Collapse of Layers
   of Grafted Polymer Chains", Macromolecules 24, 140-149.

.. [#Karim] Karim, A; Douglas, JF; Horkay, F; Fetters, LJ; Satija, SK (1996)
   "Comparative swelling of gels and polymer brush layers",
   Physica B 221, 331-336. `<http://dx.doi.org/10.1016/0921-4526(95)00946-9>`
   
.. [#Adamuţi-Trache] Adamuţi-Trache, M., McMullen, W. E. & Douglas, J. F. 
    Segmental concentration profiles of end-tethered polymers with excluded-
    volume and surface interactions. J. Chem. Phys. 105, 4798 (1996).
    
.. [#Cosgrove] Cosgrove, T., Heath, T., Van Lent, B., Leermakers, F. A. M., 
    & Scheutjens, J. M. H. M. (1987). Configuration of terminally attached 
    chains at the solid/solvent interface: self-consistent field theory and 
    a Monte Carlo model. Macromolecules, 20(7), 1692–1696. 
    doi:10.1021/ma00173a041
    
.. [#deVos] De Vos, W. M., & Leermakers, F. A. M. (2009). Modeling the 
    structure of a polydisperse polymer brush. Polymer, 50(1), 305–316. 
    doi:10.1016/j.polymer.2008.10.025
    
.. [#Sheridan] Sheridan, R. J., Beers, K. L., et. al (2014). Direct observation
    of "surface theta" conditons. {in prep}
"""
from __future__ import division
__all__ = ["PolymerBrush","PolymerMushroom","EndTetheredPolymer","VolumeProfile","layer_thickness"]
import inspect
import numpy as np
from numpy import real, imag, exp
from bumps.parameter import Parameter

from .model import Layer
from . import util


class PolymerBrush(Layer):
    r"""
    Polymer brushes in a solvent

    Parameters:

        *thickness* the thickness of the solvent layer
        *interface* the roughness of the solvent surface
        *polymer* the polymer material
        *solvent* the solvent material or vacuum
        *base_vf* volume fraction (%) of the polymer brush at the interface
        *base* the thickness of the brush interface (A)
        *length* the length of the brush above the interface (A)
        *power* the rate of brush thinning
        *sigma* rms brush roughness (A)

    The materials can either use the scattering length density directly,
    such as PDMS = SLD(0.063, 0.00006) or they can use chemical composition
    and material density such as PDMS=Material("C2H6OSi",density=0.965).

    These parameters combine in the following profile formula:

    .. math::

        V(z) &= \left\{
          \begin{array}{ll}
            V_o                        & \mbox{if } z <= z_o \\
            V_o (1 - ((z-z_o)/L)^2)^p  & \mbox{if } z_o < z < z_o + L \\
            0                          & \mbox{if } z >= z_o + L
          \end{array}
        \right. \\
        V_\sigma(z)
           &= V(z) \star
                 \frac{e^{-\frac{1}{2}(z/\sigma)^2}}{\sqrt{2\pi\sigma^2}} \\
        \rho(z) &= \rho_p V_\sigma(z) + \rho_s (1-V_\sigma(z))

    where $V_\sigma(z)$ is volume fraction convoluted with brush
    roughness $\sigma$ and $\rho(z)$ is the complex scattering
    length density of the profile.
    """
    def __init__(self, thickness=0, interface=0, name="brush",
                 polymer=None, solvent=None, base_vf=None,
                 base=None, length=None, power=None, sigma=None):
        self.thickness = Parameter.default(thickness, name="brush thickness")
        self.interface = Parameter.default(interface, name="brush interface")
        self.base_vf = Parameter.default(base_vf, name="base_vf")
        self.base = Parameter.default(base, name="base")
        self.length = Parameter.default(length, name="length")
        self.power = Parameter.default(power, name="power")
        self.sigma = Parameter.default(sigma, name="sigma")
        self.solvent = solvent
        self.polymer = polymer
        self.name = name
        # Constraints:
        #   base_vf in [0,1]
        #   base,length,sigma,thickness,interface>0
        #   base+length+3*sigma <= thickness
    def parameters(self):
        return {'solvent':self.solvent.parameters(),
                'polymer':self.polymer.parameters(),
                'base_vf':self.base_vf,
                'base':self.base,
                'length':self.length,
                'power':self.power,
                'sigma':self.sigma,
                }
                
    def profile(self, z):
        base_vf, base, length, power, sigma \
            = [p.value for p in (self.base_vf, self.base,
               self.length, self.power, self.sigma)]
        base_vf /= 100. # % to fraction
        L0 = base  # if base < thickness else thickness
        L1 = base+length # if base+length < thickness else thickness-L0
        if length == 0:
            v = np.ones_like(z)
        else:
            v = (1 - ((z-L0)/(L1-L0))**2)
        v[z<L0] = 1
        v[z>L1] = 0
        brush_profile = base_vf * v**power
        # TODO: we could use Nevot-Croce rather than smearing the profile
        vf = smear(z, brush_profile, sigma)
        return vf

    def render(self, probe, slabs):
        thickness = self.thickness.value
        Pw,Pz = slabs.microslabs(thickness)
        # Skip layer if it falls to zero thickness.  This may lead to
        # problems in the fitter, since R(thickness) is non-differentiable
        # at thickness = 0.  "Clip to boundary" range handling will at
        # least allow this point to be found.
        # TODO: consider using this behaviour on all layer types.
        if len(Pw) == 0: return

        Mr,Mi = self.polymer.sld(probe)
        Sr,Si = self.solvent.sld(probe)
        M = Mr + 1j*Mi
        S = Sr + 1j*Si
        try: M,S = M[0],S[0]  # Temporary hack
        except: pass

        vf = self.profile(Pz)
        Pw,vf = util.merge_ends(Pw, vf, tol=1e-3)
        P = M*vf + S*(1-vf)
        Pr, Pi = real(P), imag(P)
        slabs.extend(rho=[Pr], irho=[Pi], w=Pw)

def layer_thickness(z):
    """
    Return the thickness of a layer given the microslab z points.

    The z points are at the centers of the bins.  we can use the recurrence
    that boundary b[k] = z[k-1] + (z[k-1] - b[k-1]) to compute the
    total length of the layer.
    """
    return 2 * (np.sum(z[-1::-2]) - np.sum(z[-2::-2]))

class VolumeProfile(Layer):
    """
    Generic volume profile function

    Parameters:

        *thickness* the thickness of the solvent layer
        *interface* the roughness of the solvent surface
        *material* the polymer material
        *solvent* the solvent material
        *profile* the profile function, suitably parameterized

    The materials can either use the scattering length density directly,
    such as PDMS = SLD(0.063, 0.00006) or they can use chemical composition
    and material density such as PDMS=Material("C2H6OSi",density=0.965).

    These parameters combine in the following profile formula::

        sld = material.sld * profile + solvent.sld * (1 - profile)

    The profile function takes a depth z and returns a density rho.

    For volume profiles, the returned rho should be the volume fraction
    of the material.  For SLD profiles, rho should be complex scattering
    length density of the material.

    Fitting parameters are the available named arguments to the function.
    The first argument must be *z*, which is the array of depths at which
    the profile is to be evaluated.  It is guaranteed to be increasing, with
    step size 2*z[0].

    Initial values for the function parameters can be given using name=value.
    These values can be scalars or fitting parameters.  The function will
    be called with the current parameter values as arguments.  The layer
    thickness can be computed as :func:`layer_thickness`.

    """
    # TODO: test that thickness(z) matches the thickness of the layer
    def __init__(self, thickness=0, interface=0, name="VolumeProfile",
                 material=None, solvent=None, profile=None, **kw):
        if interface != 0: raise NotImplementedError("interface not yet supported")
        if profile is None or material is None or solvent is None:
            raise TypeError("Need polymer, solvent and profile")
        self.name = name
        self.thickness = Parameter.default(thickness, name="solvent thickness")
        self.interface = Parameter.default(interface, name="solvent interface")
        self.profile = profile
        self.solvent = solvent
        self.material = material

        # Query profile function for the list of arguments
        vars = inspect.getargspec(profile)[0]
        #print "vars",vars
        if inspect.ismethod(profile): vars = vars[1:]  # Chop self
        vars = vars[1:]  # Chop z
        #print vars
        unused = [k for k in kw.keys() if k not in vars]
        if len(unused) > 0:
            raise TypeError("Profile got unexpected keyword argument '%s'"%unused[0])
        dups = [k for k in vars
                if k in ('thickness','interface','polymer','solvent','profile')]
        if len(dups) > 0:
            raise TypeError("Profile has conflicting argument '%s'"%dups[0])
        for k in vars: kw.setdefault(k,0)
        for k,v in kw.items():
            setattr(self,k,Parameter.default(v,name=k))

        self._parameters = vars

    def parameters(self):
        P = {'solvent':self.solvent.parameters(),
             'material':self.material.parameters(),
             }
        for k in self._parameters:
            P[k] = getattr(self,k)
        return P

    def render(self, probe, slabs):
        Mr,Mi = self.material.sld(probe)
        Sr,Si = self.solvent.sld(probe)
        M = Mr + 1j*Mi
        S = Sr + 1j*Si
        #M,S = M[0],S[0]  # Temporary hack
        Pw,Pz = slabs.microslabs(self.thickness.value)
        if len(Pw)== 0: return
        kw = dict((k,getattr(self,k).value) for k in self._parameters)
        #print kw
        phi = self.profile(Pz,**kw)
        try:
            if phi.shape != Pz.shape: raise Exception
        except:
            raise TypeError("profile function '%s' did not return array phi(z)"
                            %self.profile.__name__)
        Pw,phi = util.merge_ends(Pw, phi, tol=1e-3)
        P = M*phi + S*(1-phi)
        slabs.extend(rho = [real(P)], irho = [imag(P)], w = Pw)
        #slabs.interface(self.interface.value)


def smear(z, P, sigma):
    """
    Gaussian smearing

    :Parameters:
        *z* | vector
            equally spaced sample times
        *P* | vector
            sample values
        *sigma* | real
            root-mean-squared convolution width
    :Returns:
        *Ps* | vector
            smeared sample values
    """
    if len(z) < 3: return P
    dz = z[1]-z[0]
    if 3*sigma < dz: return P
    w = int(3*sigma/dz)
    G = exp(-0.5*(np.arange(-w,w+1)*(dz/sigma))**2)
    full = np.hstack( ([P[0]]*w, P, [P[-1]]*w) )
    return np.convolve(full,G/np.sum(G),'valid')

class PolymerMushroom(Layer):
    """ 
    Polymer mushrooms in a solvent (volume profile)

    :Parameters:

    *delta* | interaction parameter
    *vf* | not quite volume fraction (dimensionless grafting density)
    *sigma* | convolution roughness (A)
    
    Using analytical SCF methods for gaussian chains, which are scaled
    by the radius of gyration of the equivalent free polymer as an 
    approximation to results of renormalization group methods.[#Adamuţi-Trache]
    
    Solutions are only strictly valid for vf << 1. 

.. [#Adamuţi-Trache] Adamuţi-Trache, M., McMullen, W. E. & Douglas, J. F. 
    Segmental concentration profiles of end-tethered polymers with 
    excluded-volume and surface interactions. J. Chem. Phys. 105, 4798 (1996).
    """
    
    def __init__(self, thickness=0, interface=0, name="shroom",
                 polymer=None, solvent=None, sigma=0,
                 vf=0, delta=0):
        self.thickness = Parameter.default(thickness, name="shroom thickness")
        self.interface = Parameter.default(interface, name="shroom interface")
        self.delta = Parameter.default(delta, name="delta")
        self.vf = Parameter.default(vf, name="vf")
        self.sigma = Parameter.default(sigma, name="sigma")
        self.solvent = solvent
        self.polymer = polymer
        self.name = name
    def parameters(self):
        return {'solvent':self.solvent.parameters(),
                'polymer':self.polymer.parameters(),
                'delta':self.delta,
                'vf':self.vf,
                'sigma':self.sigma,
                'thickness':self.thickness,
                'interface':self.interface
                }
                
    def profile(self, z):
        delta, sigma, vf, thickness \
            = [p.value for p in self.delta, self.sigma, self.vf, self.thickness]

        return smear(z, MushroomProfile(z, delta, vf, sigma), sigma)


    def render(self, probe, slabs):
        thickness = self.thickness.value
        Pw,Pz = slabs.microslabs(thickness)
        # Skip layer if it falls to zero thickness.  This may lead to
        # problems in the fitter, since R(thickness) is non-differentiable
        # at thickness = 0.  "Clip to boundary" range handling will at
        # least allow this point to be found.
        # TODO: consider using this behaviour on all layer types.
        if len(Pw) == 0: return

        Mr,Mi = self.polymer.sld(probe)
        Sr,Si = self.solvent.sld(probe)
        M = Mr + 1j*Mi
        S = Sr + 1j*Si
        try: M,S = M[0],S[0]  # Temporary hack
        except: pass

        phi = self.profile(Pz)
        Pw,phi = util.merge_ends(Pw, phi, tol=1e-3)
        P = M*phi + S*(1-phi)
        Pr, Pi = np.real(P), np.imag(P)
        slabs.extend(rho=[Pr], irho=[Pi], w=Pw)

from numpy import sqrt, pi, hstack, ones_like
from scipy.special import erfc, erfcx

def MushroomProfile(z, delta=0.1, vf=1.0, sigma=1.0):
    thickness = layer_thickness(z)
    thresh=1e-10
    
    base = 3.0*sigma # tail is erf, capture 95% of the mixing
    Rg = (thickness-base) / 4.0 # profile ends by ~4 RG, so we can tether these
    keep = (z-base) >= 0.0    
    x = (z[keep] - base) / Rg
    '''
    mushroom_profile_math has a divide by zero problem at delta=0.
    Fix it by weighted average of the profile above and below a threshold.
    No visual difference when delta is between +-0.001, and there's no
    floating point error until ~+-1e-14.
    '''

    if abs(delta) > thresh:
        mushroom_profile = mushroom_math(x,delta,vf)
    else: # we should RARELY get here
        scale = (delta+thresh)/2.0/thresh             
        mushroom_profile = scale*mushroom_math(x,thresh,vf)        
        mushroom_profile = (mushroom_profile +
                            (1.0-scale)*mushroom_math(x,-thresh,vf)) 

    try:
        # make the base connect with the profile
        zextra = z[np.logical_not(keep)]
        base_profile = ones_like(zextra)*mushroom_profile[0]
    except IndexError:
        base_profile = ones_like(z)*mushroom_profile[0]
        
    return hstack((base_profile, mushroom_profile))

sqrt_pi=sqrt(pi)
def mushroom_math(x,delta=.1,vf=.1):
    '''
    new method, rewrite for numerical stability at high delta
    delta=0 causes divide by zero error!! Compensate elsewhere.
    http://ab-initio.mit.edu/wiki/index.php/Faddeeva_Package
    '''
    
    return (
            (
             erfc(x/2.0) 
             -erfcx(2.0*(delta+x/4.0))*exp(-((x/2.0)**2)) 
             -erfc(x)
             + (
                (1.0-4.0*delta*(x+2.0*delta))*erfcx(2.0*(delta+x/2.0))
                + 4.0*delta/sqrt_pi
               ) * exp(-(x**2))
            ) * vf/(2.0 * delta * erfcx(2.0*delta))
           )

class EndTetheredPolymer(Layer):
    """
    Polymer end-tethered to an interface in a solvent
    
    :Parameters:

    *z* | the SLD depth array
    *chi* | solvent interaction parameter
    *chi_s* | surface interaction parameter
    *h_dry* | thickness of the neat polymer layer
    *l_lat* | real length per lattice site
    *mn* | Number average molecular weight
    *m_lat* | real mass per lattice segment
    *pdi* | Dispersity (Polydispersity index)
    *thickness* | Slab thickness should be greater than the contour length
            of the polymer
    *interface* | should be zero
    *material* | the polymer material
    *solvent* | the solvent material
    
    Previous layer should not have roughness! Use a spline to simulate it. 
    You can put this material in the "above" slot of Freelayer.
    
    According to [#Vincent], l_lat and m_lat should be calculated by the formulas
    
    l_lat = a**2 * m / l / p_l
    m_lat = (a * m / l)**2 / p_l
    
    where l is the real polymer's bond length, m is the real segment mass,
    and a is the proportionality constant between molecular weight and radius
    of gyration at theta conditions. The lattice persistence is 
    
    p_l = 1/6*((1+1/Z)/(1-1/Z))
    
    with coordination number Z = 6 for a cubic lattice, p_l = .233.

..  [#Vincent] Vincent, B., Edwards, J., Emmett, S., & Croot, R. (1988). 
        Phase separation in dispersions of weakly-interacting particles in 
        solutions of non-adsorbing polymer. Colloids and Surfaces, 31, 267–298.
        doi:10.1016/0166-6622(88)80200-2

    """
    
    def __init__(self, thickness=0, interface=0, name="EndTetheredPolymer",
                 polymer=None, solvent=None, chi=0, chi_s=0, h_dry=None, 
                 l_lat=1, mn=None, m_lat=1, pdi=1):
        if interface != 0: raise NotImplementedError("interface not yet supported")
        if polymer is None or solvent is None or h_dry is None or mn is None:
            raise TypeError("Need polymer, solvent and profile")
        
        self.thickness = Parameter.default(thickness, name="SCF thickness")
        self.interface = Parameter.default(interface, name="SCF interface")
        self.chi   = Parameter.default(chi, name="chi")
        self.chi_s = Parameter.default(chi_s, name="surface chi")
        self.h_dry = Parameter.default(h_dry, name="dry thickness")
        self.l_lat = Parameter.default(l_lat, name="lattice layer length")
        self.mn    = Parameter.default(mn, name="Num. Avg. MW")
        self.m_lat = Parameter.default(m_lat, name="lattice segment mass")
        self.pdi   = Parameter.default(pdi, name="Dispersity")
        self.phi_prev = None
        self.solvent = solvent
        self.polymer = polymer
        self.name = name
        
    def parameters(self):
        return {'solvent':self.solvent.parameters(),
                'polymer':self.polymer.parameters(),
                'chi':self.chi,
                'chi_s':self.chi_s,
                'h_dry':self.h_dry,
                'l_lat':self.l_lat,
                'mn':self.mn,
                'm_lat':self.m_lat,
                'pdi':self.pdi,
                'thickness':self.thickness,
                'interface':self.interface
                }
                
    def profile(self, z):
        phi = SCFprofile(z, chi=self.chi.value, chi_s=self.chi_s.value, 
                 h_dry=self.h_dry.value,l_lat=self.l_lat.value,
                 mn=self.mn.value, m_lat=self.m_lat.value, pdi=self.pdi.value,
                 phi0=self.phi_prev)
        self.phi_prev = phi.copy()
        return phi

    def render(self, probe, slabs):
        thickness = self.thickness.value
        Pw,Pz = slabs.microslabs(thickness)
        # Skip layer if it falls to zero thickness.  This may lead to
        # problems in the fitter, since R(thickness) is non-differentiable
        # at thickness = 0.  "Clip to boundary" range handling will at
        # least allow this point to be found.
        # TODO: consider using this behaviour on all layer types.
        if len(Pw) == 0: return

        Mr,Mi = self.polymer.sld(probe)
        Sr,Si = self.solvent.sld(probe)
        M = Mr + 1j*Mi
        S = Sr + 1j*Si
        try: M,S = M[0],S[0]  # Temporary hack
        except: pass

        phi = self.profile(Pz)
        Pw,phi = util.merge_ends(Pw, phi, tol=1e-3)
        P = M*phi + S*(1-phi)
        Pr, Pi = np.real(P), np.imag(P)
        slabs.extend(rho=[Pr], irho=[Pi], w=Pw)
    
from numpy import absolute as abs
import time
from numpy.linalg import norm
from scipy.optimize import  root

MINLAT = 35

def SCFprofile(z, chi=None, chi_s=None, h_dry=None, l_lat=1, mn=None, 
               m_lat=1, pdi=1, phi0=None, disp=None):
    ''' Generate volume fraction profile for Refl1D based on real parameters.
    
    The field theory is a lattice-based one, so we need to move between lattice
    and real space. This is done using the parameters l_lat and m_lat, the 
    lattice size and the mass of a lattice segment, respectivley. We use h_dry
    (dry thickness) as a convenient measure of surface coverage, along with mn
    (number average molecular weight) as the real inputs.
    
    Make sure your inputs for h_dry/l_lat and mn/m_lat match dimensions! 
    Angstroms and daltons are good choices.
    
    This function is suitable for use as a VolumeProfile, as well as the 
    default EndTetheredPolymer class.
    '''
    
    # calculate lattice space parameters    
    theta = h_dry/l_lat
    r = int(mn/m_lat-.5)
    
    if phi0 is not None:
        # squeeze or stretch our initial guess to match the input space
        phi0_len = len(phi0)
        z_len = len(z)
        if phi0_len < z_len:
            phi0 = hstack((phi0,np.zeros(z_len-phi0_len)))
        elif phi0_len > z_len:
            phi0 = phi0[0:z_len]
            
        # determine number of lattice layers needed
        keep = phi0/theta > 1e-6
        z_end = z[keep].max()
        layers = round(z_end/l_lat-0.5)
        z_end = l_lat*(layers+.5) # match rounding exactly
        
        # convert initial guess to lattice space
        z_lat = np.linspace(l_lat/2,z_end,num=layers)
        phi0 = np.interp(z_lat,z,phi0)

    # solve the self consistent field equations    
    phi_lat = SCFsolve(chi=chi,chi_s=chi_s,pdi=pdi,
                     theta=theta,r=r,disp=disp,phi0=phi0)
    if disp: print "lattice segments: ", r
    
    # re-dimensionalize the solution
    layers = len(phi_lat)
    z_end = l_lat*(layers+.5)
    keep = z<=z_end
    z_lat = np.linspace(l_lat/2,z_end,num=layers)
    phi = np.interp(z[keep],z_lat,phi_lat)

    # fill in the end with zeros
    zextra = z[np.logical_not(keep)]
    
    return hstack((phi,np.zeros_like(zextra)))
    
def SCFsolve(chi=0,chi_s=0,pdi=1,theta=None,r=None,disp=False,phi0=None):
    ''' Solve SCF equations using an initial guess and lattice parameters
    
    This function checks which special case self consistent field equations
    we are using, evaluates any starting guess against a simple default, then
    finds a solution for the equations where the lattice size is sufficiently
    large.
    
    The Newton-Krylov solver really makes this one. krylov+gmres was faster
    than the other scipy.optimize alternatives by quite a lot.
    '''
    
    sigmainput = theta/r

    if sigmainput >= 1:
        raise ValueError('Chains that short cannot be squeezed that high')
    
    if pdi==1:
        def SCFeqns(phi):
            return SCFeqns_Cosgrove(phi,chi,chi_s,theta,pdi,r)
    elif pdi>1:
        try:
            p_i = SZdist(pdi,r)
            def SCFeqns(phi):
                return SCFeqns_deVos(phi,chi,chi_s,theta,pdi,r,p_i)
        except SZerror:
            if disp: print 'PDI below threshold of numerical stability. Defaulting to calculations of uniform polymer.'
            def SCFeqns(phi):
                return SCFeqns_Cosgrove(phi,chi,chi_s,theta,pdi,r)
    else:
        raise ValueError('Invalid PDI')
    
    starttime = time.time()   
    
    default_layers = round(max(MINLAT,theta/sqrt(sigmainput)))
    default_phi0 = np.linspace(sqrt(sigmainput),0,num=default_layers)
    default_phi0 = default_phi0.ravel()

    # Check if default guess is a better guess than input
    if phi0 is None:
        phi0 = default_phi0
        layers = default_layers
        if disp: print '\nno guess passed, using default phi0: layers =',layers,'\n'
    else:
        phi0 = abs(phi0)
        phi0[phi0>1.0] = 1.0
        layers = len(phi0)
        eps = SCFeqns(phi0)
        default_eps = SCFeqns(default_phi0)
        if norm(eps)/sqrt(layers) > norm(default_eps)/sqrt(default_layers):
            if disp: print "\ndefault phi0 is better: layers =",default_layers,'\n'
            phi0 = default_phi0
        else:
            if disp: print "\npassed phi0 is better: layers =", layers,'\n'
    
    done = False
    tol = 2e-6*theta # We tolerate up to 2ppm of our polymer in the last layer
    ratio = 1.2 # otherwise we grow it by 20%
    layers = len(phi0)
    # if the loop sees that it can shrink 20%, it will, but it can lead to
    # endless loops. This flag shows if it has grown before, so it knows to
    # quit instead of shrinking
    growing = False
    
    # apparently scope rules dictate that we can change 'layers' without a
    # redefinition of this callback, so i got rid of the one in the while block    
    def callback(x,*args,**kwargs): 
        _proto_callback(x,disp,layers,tol,ratio)
    
    while not done:
        if disp: print "\nSolving SCF equation set..."
        try:
            result = root(SCFeqns,phi0,method='Krylov',callback=callback,
                    options={'disp':bool(disp),
                             'jac_options':{'method':'gmres'}})
            phi = abs(result.x)
            if disp: print '\nSolver exit code:',result.status,result.message
        except ShortCircuitError as e:
            growing = False # scrub this flag so we don't quit directly (rare)
            phi = e.x
            if disp: print e
            
        if disp: print 'phi(L)/sum(phi) =',phi[-1] / theta * 1e6,'(ppm)\n'
        
        if phi[-1] > tol:
            # if the last layer is beyond tolerance, grow the lattice
            newlayers = max(1,round(layers*(ratio-1)))
            if disp: print 'Growing undersized lattice by', newlayers
            phi0 = np.append(phi,np.linspace(phi[-1],0,num=newlayers))
            growing = True
        elif layers > MINLAT and phi[round(layers/ratio)] < tol:
            # if the layer at 83% of the thickenss is less than the tolerance,
            # we can shrink it, but not if it's already a small lattice, or 
            # if we grew after the last successful call of root()
            if growing:
                done = True
            else:
                if disp: print 'Shrinking undersized lattice...'
                phi0 = phi[0:round(layers/ratio)]
        else:
            # otherwise, we are done for real
            done = True
            
        layers = len(phi0)
            
    if disp:
        print "execution time:", round(time.time()-starttime,3), "s"
        print "lattice size:", layers
    
    return phi
    
class ShortCircuitError(Exception):
    ''' Special error to stop root() before a solution is found.
    
    '''
    def __init__(self, value,x):
         self.value = value
         self.x = x
    def __str__(self):
         return repr(self.value)
         
def _proto_callback(x,disp,layers,tol,ratio):
    ''' Special callback to stop root() before solution is found.
    
    This kills root if the tolerances are exceeded by 4 times the tolerances
    of the lattice resizing loop. This seems to work well empirically to 
    restart the solver when necessary without cutting out of otherwise 
    reasonable solver trajectories.
    '''
    if disp: print "Iterating..."
    if x[-1] > 4*tol:
        raise ShortCircuitError('Stopping, lattice too small',x)
    elif layers > MINLAT and x[min(layers-1,round(layers/ratio))] < 4*tol:
        raise ShortCircuitError('Stopping, lattice too big',x)

from scipy.special import gammaln

lambda_1 = np.float64(1.0)/6.0 #always assume cubic lattice (1/6) for now
lambda_0 = 1.0-2.0*lambda_1
lambda_array = np.array([lambda_1,lambda_0,lambda_1])
def _fzeros(*args):
    return np.zeros(args,dtype=np.float64,order='F')

def SCFeqns_deVos(phi_z,chi,chi_s,theta,pdi,navgsegments,p_i,
                  disp=False,fulloutput=False):
    ''' System of SCF equation for disperse terminally attached polymers.
    
        Formatted for input to a nonlinear minimizer or solver.
    '''
    
    # let the solver go negative if it wants, or outside the limits 0..1
    phi_z = abs(phi_z.ravel())
    if np.any(phi_z>=1):
        return np.empty_like(phi_z).fill(np.nan)
    phi_z_0 = 1.0 - phi_z
    
    layers = phi_z.size
    cutoff = p_i.size
    sigma = theta/navgsegments
    
    if pdi <= 1.0:
        raise ValueError('de Vos nSCF equations only valid for disperse polymer')
    
    # calculate all needed quantities for new g_z
    delta = _fzeros(layers)
    delta[0] = 1.0
    phi_z_avg = calc_phi_z_avg(phi_z)
    phi_z_0_avg = calc_phi_z_avg(phi_z_0)
    
    # calculate new g_z
    g_z = phi_z_0*exp(chi*(phi_z_avg-phi_z_0_avg) + chi_s*delta)
    
    # normalize g_z for numerical stability
    u = -np.log(g_z)
    uavg = np.mean(u)
    g_z_norm = g_z*exp(uavg)
    
    # calculate needed quantites for new phi_z
    g_zs_ta_norm = calc_g_zs(g_z_norm,1,layers,cutoff)
    c_i_norm = sigma*p_i/np.sum(g_zs_ta_norm,axis=0)
    g_zs_free_ngts_norm = calc_g_zs(g_z_norm,c_i_norm,layers,cutoff)
    
    phi_z_new = calc_phi_z(g_zs_ta_norm,g_zs_free_ngts_norm,g_z_norm)
    eps_z = phi_z-phi_z_new
    
    return eps_z
    
def SCFeqns_Cosgrove(phi_z,chi,chi_s,theta,pdi,segments,
                     disp=False,fulloutput=False):
    ''' System of SCF equation for uniform terminally attached polymers.
    
        Formatted for input to a nonlinear minimizer or solver.
    '''
    
    # let the solver go negative if it wants, or outside the limits 0..1
    phi_z = abs(phi_z.ravel()) 
    if np.any(phi_z>=1):
        return np.empty_like(phi_z).fill(np.nan)
    phi_z_0 = 1.0 - phi_z
    layers = phi_z.size
    
    if pdi != 1.0 and disp:
        print "Cosgrove nSCF assume PDI == 1"
        
    # calculate all needed quantities for new g_z
    delta = _fzeros(layers)
    delta[0] = 1.0
    phi_z_avg = calc_phi_z_avg(phi_z)
    phi_z_0_avg = calc_phi_z_avg(phi_z_0)
    
    # calculate new g_z and difference given all inputs
    g_z = phi_z_0*exp(chi*(phi_z_avg-phi_z_0_avg)+chi_s*delta)
    u = -np.log(g_z)
    uavg = np.mean(u)
    g_z_norm = g_z*exp(uavg)
    
    # calculate needed quantites for new phi_z
    g_ta_norm = calc_g_zs(g_z_norm,1,layers,segments)
    sigma = theta/segments
    p_i = _fzeros(1,segments)
    p_i[0,-1] = 1
    c_i_norm = sigma*p_i/np.sum(g_ta_norm[:,segments-1])
    g_free_norm = calc_g_zs(g_z_norm,c_i_norm,layers,segments)
    
    phi_z_new = calc_phi_z(g_ta_norm,g_free_norm,g_z_norm)
    eps_z = phi_z-phi_z_new
    
    return eps_z

def calc_phi_z_avg(phi_z):
    return np.convolve(phi_z,lambda_array,'same')
    
def calc_phi_z(g_ta,g_free,g_z):
    return np.sum(g_ta*np.fliplr(g_free),axis=1)/g_z

def calc_g_zs(g_z,c_i,layers,segments):
    from refl1d.calc_g_zs_cex import _calc_g_zs_inner
    
    # initialize     

    g_zs=np.empty((layers,segments),dtype=np.float64,order='F')
    
    # choose special case 
    
    if np.size(c_i) == 1:
        # terminally attached chains
        c_i = _fzeros(1,segments)
        g_zs[:,0] = _fzeros(layers)
        g_zs[0,0] = g_z[0]
    else:
        # free chains
        g_zs[:,0] = c_i[0,segments-1]*g_z
    
    # inner loops
    
    # FASTEST: call some custom C code identical to "SLOW" loop
        # beware, this changes g_zs _in_place!_
    _calc_g_zs_inner(g_z,c_i,g_zs,lambda_0,lambda_1,layers,segments)
    
    # FASTER: use the convolve function to partially vectorize  
#    pg_zs=g_zs[:,0]    
#    for r in range(1,segments):
#        pg_zs=g_z*(c_i[0,segments-r-1]+np.convolve(pg_zs,lambda_array,'same'))
#        g_zs[:,r]=pg_zs
    
    # SLOW: loop outright, pulling some slicing out of the innermost loop  
#    pg_zs=g_zs[:,0] 
#    for r in range(1,segments):
#        c=c_i[0,segments-r-1]
#        z=0
#        g_zs[z,r]=(pg_zs[z]*lambda_0+pg_zs[z+1]*lambda_1+c)*g_z[z]
#        for z in range(1,(layers-1)):
#            g_zs[z,r]=(pg_zs[z-1]*lambda_1+pg_zs[z]*lambda_0+pg_zs[z+1]*lambda_1+c)*g_z[z]
#        z=layers-1
#        g_zs[z,r]=(pg_zs[z]*lambda_0+pg_zs[z-1]*lambda_1+c)*g_z[z]
#        pg_zs=g_zs[:,r]
               
    return g_zs
    
def SZdist(pdi,nn):
    ''' Calculate Shultz-Zimm distribution from PDI and number average DP
    
    Shultz-Zimm is a "realistic" distribution for linear polymers. Numerical
    problems arise when the distribution gets too uniform, so we raise an error
    which can be caught to default to an exact uniform calculation.
    '''
    
    if pdi <= 1.0:
        raise ValueError('Invalid PDI')
        
    x = 1.0/(pdi-1.0)
    cutoff = 9000
    
    ni = np.arange(1,cutoff+1,dtype=np.float64)
    r = ni/nn
    p_ni = exp(
                  np.log(x/ni) - gammaln(x+1) + x*r*(np.log(x*r)/r-1)
                  )
    
    if any(p_ni>=1):
        raise SZerror('Schultz-Zimm calculation blew up')
    
    mysums = np.cumsum(p_ni)
    keep = np.logical_and(
        np.logical_or(np.array(range(cutoff)) < nn, p_ni >= 1.0e-6),
        mysums < 1)
        
    return p_ni[keep].reshape(1,-1)
    
class SZerror(Exception):
     def __init__(self, value):
         self.value = value
     def __str__(self):
         return repr(self.value)



#!/usr/bin/env python
import sys
import os

if len(sys.argv) == 1:
    sys.argv.append('install')

# Use our own nose-based test harness
if sys.argv[1] == 'test':
    from subprocess import call
    sys.exit(call([sys.executable, 'test.py']+sys.argv[2:]))

#sys.dont_write_bytecode = True

#from distutils.core import Extension
from setuptools import setup, find_packages, Extension
#import fix_setuptools_chmod

# Grab the openmp extension builder from the bumps package
try:
    import bumps
except:
    from os.path import dirname, abspath, join as joinpath
    bumps_path = joinpath(dirname(dirname(abspath(__file__))),'bumps')
    sys.path.insert(0, bumps_path)
from bumps.openmp_ext import openmp_ext

import refl1d


packages = find_packages()

# reflmodule extension
def reflmodule_config():
    S = ("reflmodule.cc","methods.cc",
         "reflectivity.cc", "magnetic.cc",
         "contract_profile.cc",
         "convolve.c", "convolve_sampled.c",
    )
    Sdeps = ("erf.c","methods.h","rebin.h","rebin2D.h","reflcalc.h")
    sources = [os.path.join('refl1d','lib',f) for f in S]
    depends = [os.path.join('refl1d','lib',f) for f in Sdeps]
    module = Extension('refl1d.reflmodule',
                       sources=sources, depends=depends)
    return module

from numpy import get_include
# SCF extension
def SCFmodule_config():
    return Extension('refl1d.calc_g_zs_cex', sources=[os.path.join('refl1d','lib','calc_g_zs_cex.c')],
                     include_dirs=[get_include()])
                     
#TODO: write a proper dependency checker for packages which cannot be
# installed by easy_install
#dependency_check('numpy>=1.0', 'scipy>=0.6', 'matplotlib>=1.0', 'wx>=2.8.9')

#sys.dont_write_bytecode = False
dist = setup(
        name = 'refl1d',
        version = refl1d.__version__,
        author='Paul Kienzle',
        author_email='pkienzle@nist.gov',
        url='http://www.reflectometry.org/danse/model1d.html',
        description='1-D reflectometry modelling',
        long_description=open('README.txt').read(),
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'License :: Public Domain',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Scientific/Engineering :: Physics',
            ],
        packages = packages,
        #package_data = gui_resources.package_data(),
        scripts = ['bin/refl1d_cli.py','bin/refl1d_gui.py'],
        ext_modules = [reflmodule_config(),SCFmodule_config()],
        install_requires = ['bumps>=0.7.5','periodictable'],
        cmdclass = {'build_ext': openmp_ext(default=False)},
        )

# End of file



# This program is public domain
# Author: Paul Kienzle
"""
Magnetic modeling for 1-D reflectometry.

Magnetic properties are tied to the structural description of the
but only loosely.

There may be dead regions near the interfaces of magnetic materials.

Magnetic behaviour may be varying in complex ways within and
across structural boundaries.  For example, the ma
Indeed, the pattern may continue
across spacer layers, going to zero in the magnetically dead
region and returning to its long range variation on entry to
the next magnetic layer.  Magnetic multilayers may exhibit complex
magnetism throughout the repeated section while the structural
components are fixed.

The scattering behaviour is dependent upon net field strength relative to
polarization direction.   This arises from three underlying quantities:
the strength of the individual dipole moments in the layer, the degree
of alignment of these moments, and the net direction of the alignment.
The strength of the dipole moment depends on the details of the electronic
structure, so unlike the nuclear scattering potential, it cannot be readily
determined from material composition.  Similarly, net magnetization
depends on the details of the magnetic domains within the material, and
cannot readily be determined from first principles.  The interaction
potential of the net magnetic moment depends on the alignment of the field
with respect to the beam, with a net scattering length density of
:math:`\rho_M \cos(\theta_M)`.  Clearly the scattering measurement will
not be able to distinguish between a reduced net magnetic strength
:math:`\rho_M` and a change in orientation :math:`\theta_M` for an
individual measurement, as should be apparent from the correlated
uncertainty plot produced when both parameters are fit.

Magnetism support is split into two parts: describing the layers
and anchoring them to the structure.
"""
import numpy
from numpy import inf
from bumps.parameter import Parameter, flatten
from bumps.mono import monospline

from .model import Layer, Stack

class BaseMagnetism(object):
    """
    Magnetic properties of the layer.

    Magnetism is attached to set of nuclear layers by setting the
    *magnetism* property of the first layer to the rendered for the
    magnetic profile, and setting the *magnetism.extent* property to
    say how many layers it extends over.

    *dead_below* and *dead_above* are dead regions within the magnetic
    extent, which allow you to shift the magnetic interfaces relative
    to the nuclear interfaces.

    *interface_below* and *interface_above* are the interface widths
    for the magnetic layer, which default to the interface widths for
    the corresponding nuclear layers if no interfaces are specified.
    For consecutive layers, only *interface_above* is used; any value
    for *interface_below* is ignored.
    """
    def __init__(self, extent=1,
                 dead_below=0, dead_above=0,
                 interface_below=None, interface_above=None,
                 name="LAYER"):
        self.dead_below = Parameter.default(dead_below, limits=(0,inf),
                                            name=name+" deadM below")
        self.dead_above = Parameter.default(dead_above, limits=(0,inf),
                                            name=name+" deadM above")
        if interface_below is not None:
            interface_below = Parameter.default(interface_below, limits=(0,inf),
                                  name=name+" interfaceM below")
        if interface_above is not None:
            interface_above = Parameter.default(interface_above, limits=(0,inf),
                                  name=name+"  interfaceM above")
        self.interface_below = interface_below
        self.interface_above = interface_above
        self.name = name
        self.extent = extent
    def parameters(self):
        return {'dead_below':self.dead_below,
                'dead_above':self.dead_above,
                'interface_below':self.interface_below,
                'interface_above':self.interface_above,
                }

    def set_layer_name(self, name):
        """
        Update the names of the magnetic parameters with the name of the
        layer if it has not already been set.  This is necessary since we don't
        know the layer name until after we have constructed the magnetism object.
        """
        if self.name == "LAYER":
            for p in flatten(self.parameters()):
                p.name = p.name.replace("LAYER",name)
            self.name = name

class Magnetism(BaseMagnetism):
    """
    Region of constant magnetism.
    """
    def __init__(self, rhoM=0, thetaM=270, name="LAYER", **kw):
        BaseMagnetism.__init__(self, name=name, **kw)
        self.rhoM = Parameter.default(rhoM, name=name+" rhoM")
        self.thetaM = Parameter.default(thetaM, limits=(0,360),
                                        name=name+" thetaM")

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM)
        return parameters

    def render(self, probe, slabs, thickness, anchor, sigma):
        slabs.add_magnetism(anchor=anchor,
                            w=[thickness],
                            rhoM=[self.rhoM.value],
                            thetaM=[self.thetaM.value],
                            sigma=sigma)
    def __str__(self):
        return "magnetism(%g)"%self.rhoM.value
    def __repr__(self):
        return ("Magnetism(rhoM=%g,thetaM=%g)"
                %(self.rhoM.value,self.thetaM.value))

class MagnetismStack(BaseMagnetism):
    """
    Magnetic slabs within a magnetic layer.
    """
    def __init__(self, weight=[], rhoM=[], thetaM=[270], interfaceM=[0],
                 name="LAYER", **kw):
        if (len(thetaM) != 1 and len(thetaM) != len(weight)
            and len(rhoM) != 1 and len(rhoM) != len(weight)
            and len(interfaceM) != 1 and len(interfaceM) != len(weight)-1):
            raise ValueError("Must have one rhoM, thetaM and intefaceM for each layer")
        if interfaceM != [0]:
            raise NotImplementedError("Doesn't support magnetic roughness")

        BaseMagnetism.__init__(self, stack=stack, name=name, **kw)
        self.weight = [Parameter.default(v, name=name+" weight[%d]"%i)
                       for i,v in enumerate(weight)]
        self.rhoM = [Parameter.default(v, name=name+" rhoM[%d]"%i)
                     for i,v in enumerate(rhoM)]
        self.thetaM = [Parameter.default(v, name=name+" thetaM[%d]"%i)
                       for i,v in enumerate(thetaM)]
        self.interfaceM = [Parameter.default(v, name=name+" interfaceM[%d]"%i)
                           for i,v in enumerate(interfaceM)]

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(rhoM=self.rhoM,
                          thetaM=self.thetaM,
                          interfaceM=self.interfaceM,
                          weight=self.weight)
        return parameters

    def render(self, probe, slabs, thickness, anchor, sigma):
        w = numpy.array([p.value for p in self.weight])
        w *= thickness / numpy.sum(w)
        rhoM = [p.value for p in self.rhoM]
        thetaM = [p.value for p in self.thetaM]
        sigmaM = [p.value for p in self.interfaceM]
        if len(rhoM) == 1: rhoM = [rhoM[0]]*len(w)
        if len(thetaM) == 1: thetaM = [thetaM[0]]*len(w)
        if len(sigmaM) == 1: sigmaM = [sigmaM[0]]*(len(w)-1)

        slabs.add_magnetism(anchor=anchor,
                            w=w,rhoM=rhoM,thetaM=thetaM,
                            sigma=sigma)

    def __str__(self):
        return "MagnetismStack(%d)"%(len(self.rhoM))
    def __repr__(self):
        return "MagnetismStack"


class MagnetismTwist(BaseMagnetism):
    """
    Linear change in magnetism throughout layer.
    """
    magnetic = True
    def __init__(self, 
                 rhoM=[0,0], thetaM=[270,270],
                 name="LAYER", **kw):
        BaseMagnetism.__init__(self, name=name, **kw)
        self.rhoM = [Parameter.default(v, name=name+" rhoM[%d]"%i)
                     for i,v in enumerate(rhoM)]
        self.thetaM = [Parameter.default(v, name=name+" thetaM[%d]"%i)
                       for i,v in enumerate(thetaM)]

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM)
        return parameters

    def render(self, probe, slabs, thickness, anchor, sigma):
        w,z = slabs.microslabs(thickness)
        rhoM = numpy.linspace(self.rhoM[0].value,
                              self.rhoM[1].value,len(z))
        thetaM = numpy.linspace(self.thetaM[0].value,
                                self.thetaM[1].value,len(z))
        slabs.add_magnetism(anchor=anchor,
                            w=w,rhoM=rhoM,thetaM=thetaM,
                            sigma=sigma)

    def __str__(self):
        return "twist(%g->%g)"%(self.rhoM[0].value,self.rhoM[1].value)
    def __repr__(self):
        return "MagneticTwist"


class FreeMagnetism(BaseMagnetism):
    """
    Spline change in magnetism throughout layer.
    """
    magnetic = True
    def __init__(self, z = [], rhoM = [], thetaM = [],
                 name="LAYER", **kw):
        BaseMagnetism.__init__(self, name=name, **kw)
        def parvec(vector,name,limits):
            return [Parameter.default(p,name=name+"[%d]"%i,limits=limits)
                    for i,p in enumerate(vector)]
        self.rhoM, self.thetaM, self.z \
            = [parvec(v,name+" "+part,limits)
               for v,part,limits in zip((rhoM, thetaM, z),
                                        ('rhoM', 'thetaM', 'z'),
                                        ((0,inf),(0,360),(0,1))
                                        )]
        if len(self.z) != len(self.rhoM):
            raise ValueError("must have number of intervals dz one less than rhoM")
        if len(self.thetaM) > 0 and len(self.rhoM) != len(self.thetaM):
            raise ValueError("must have one thetaM for each rhoM")

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(rhoM=self.rhoM, thetaM=self.thetaM, z=self.z)
        return parameters

    def profile(self, Pz, thickness):
        mbelow,tbelow = 0,(self.thetaM[0].value if self.thetaM else 270)
        mabove,tabove = 0,(self.thetaM[-1].value if self.thetaM else 270)
        z = numpy.sort([0]+[p.value for p in self.z]+[1])*thickness

        rhoM = numpy.hstack((mbelow, [p.value for p in self.rhoM], mabove))
        PrhoM = monospline(z, rhoM, Pz)

        if numpy.any(numpy.isnan(PrhoM)):
            print("in mono with bad PrhoM")
            print("z %s"%str(z))
            print("p %s"%str([p.value for p in self.z]))


        if len(self.thetaM)>1:
            thetaM = numpy.hstack((tbelow, [p.value for p in self.thetaM], tabove))
            PthetaM = monospline(z, thetaM, Pz)
        elif len(self.thetaM)==1:
            PthetaM = self.thetaM.value * numpy.ones_like(PrhoM)
        else:
            PthetaM = 270*numpy.ones_like(PrhoM)
        return PrhoM,PthetaM

    def render(self, probe, slabs, thickness, anchor, sigma):
        Pw,Pz = slabs.microslabs(thickness)
        rhoM,thetaM = self.profile(Pz, thickness)
        slabs.add_magnetism(anchor=anchor,
                            w=Pw,rhoM=rhoM,thetaM=thetaM,
                            sigma=sigma)

    def __str__(self):
        return "freemag(%d)"%(len(self.rhoM))
    def __repr__(self):
        return "FreeMagnetism"
